import argparse
import sys
import sysconfig

from pdm import termui
from pdm.cli import actions
from pdm.cli.commands.base import BaseCommand
from pdm.cli.filters import GroupSelection
from pdm.cli.hooks import HookManager
from pdm.cli.options import (
    dry_run_option,
    frozen_lockfile_option,
    groups_group,
    install_group,
    lockfile_option,
    override_option,
    skip_option,
    venv_option,
)
from pdm.project import Project


class Command(BaseCommand):
    """Install dependencies from lock file"""

    arguments = (
        *BaseCommand.arguments,
        groups_group,
        install_group,
        override_option,
        dry_run_option,
        lockfile_option,
        frozen_lockfile_option,
        skip_option,
        venv_option,
    )

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "--check",
            action="store_true",
            help="Check if the lock file is up to date and fail otherwise",
        )
        parser.add_argument("--plugins", action="store_true", help="Install the plugins specified in pyproject.toml")

    def install_plugins(self, project: Project) -> None:
        from pdm.environments import PythonEnvironment
        from pdm.installers.core import install_requirements
        from pdm.models.requirements import parse_line

        plugins = [parse_line(r) for r in project.pyproject.plugins]
        if not plugins:
            return
        plugin_root = project.root / ".pdm-plugins"
        extra_paths = list({sysconfig.get_path("purelib"), sysconfig.get_path("platlib")})
        environment = PythonEnvironment(
            project, python=sys.executable, prefix=str(plugin_root), extra_paths=extra_paths
        )
        with project.core.ui.open_spinner("[success]Installing plugins...[/]"):
            with project.core.ui.logging("install-plugins"):
                install_requirements(
                    plugins, environment, clean=True, use_install_cache=project.config["install.cache"], allow_uv=False
                )
            if not plugin_root.joinpath(".gitignore").exists():
                plugin_root.mkdir(exist_ok=True)
                plugin_root.joinpath(".gitignore").write_text("*\n")
        project.core.ui.echo("Plugins are installed successfully into [primary].pdm-plugins[/].")

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        if not project.pyproject.is_valid and termui.is_interactive():
            actions.ask_for_import(project)

        if options.plugins:
            return self.install_plugins(project)

        hooks = HookManager(project, options.skip)

        strategy = actions.check_lockfile(project, False)
        selection = GroupSelection.from_options(project, options)
        if strategy:
            if options.check:
                project.core.ui.echo(
                    "Please run [success]`pdm lock`[/] to update the lock file",
                    err=True,
                )
                sys.exit(1)
            if project.enable_write_lockfile:
                project.core.ui.echo("Updating the lock file...", style="success", err=True)
            # We would like to keep the selected groups when the lockfile exists
            # but use the groups passed-in when creating a new lockfile or doing a non-lock install.
            if strategy == "all" or not project.enable_write_lockfile:
                lock_selection = selection
            else:
                lock_selection = GroupSelection(project)
            actions.do_lock(
                project,
                strategy=strategy,
                dry_run=options.dry_run,
                hooks=hooks,
                groups=lock_selection.all(),
            )

        actions.do_sync(
            project,
            selection=selection,
            no_editable=options.no_editable,
            no_self=options.no_self or "default" not in selection,
            dry_run=options.dry_run,
            fail_fast=options.fail_fast,
            hooks=hooks,
        )



from __future__ import annotations

from copy import deepcopy
from pathlib import Path
from typing import Any, cast

import tomlkit

from pdm import termui
from pdm.compat import tomllib


class TOMLFile:
    def __init__(self, path: str | Path, *, parse: bool = True, ui: termui.UI) -> None:
        from tomlkit.toml_file import TOMLFile as TomlkitTOMLFile

        self._file = TomlkitTOMLFile(path)
        self.ui = ui
        self._data = self._parse() if parse else {}
        self._for_write = False

    @property
    def _path(self) -> Path:
        return Path(self._file._path)

    def _parse(self) -> dict[str, Any]:
        # By default, use tomllib for parsing as it is much faster
        try:
            with open(self._path, "rb") as fp:
                return tomllib.load(fp)
        except FileNotFoundError:
            return {}

    def open_for_write(self) -> tomlkit.TOMLDocument:
        # Ensure the document is re-parsed by tomlkit for writing with styles preserved
        if self._for_write:
            return cast(tomlkit.TOMLDocument, self._data)
        try:
            self._data = self._file.read()
        except FileNotFoundError:
            self._data = tomlkit.document()
        self._for_write = True
        return self._data

    def open_for_read(self) -> dict[str, Any]:
        """Get the (read-only) data of the TOML file."""
        if hasattr(self._data, "unwrap"):
            return self._data.unwrap()  # type: ignore[attr-defined]
        return deepcopy(self._data)

    def set_data(self, data: dict[str, Any]) -> None:
        """Set the data of the TOML file."""
        self._data = data
        self._for_write = True

    def reload(self) -> None:
        self._data = self._parse()
        self._for_write = False

    def write(self) -> None:
        if not self._for_write:
            raise RuntimeError("TOMLFile not opened for write. Call open_for_write() first.")
        self._path.parent.mkdir(parents=True, exist_ok=True)
        if isinstance(self._data, tomlkit.TOMLDocument):
            data = self._data
        else:
            data = tomlkit.document()
            data.update(self._data)
        self._file.write(data)

    def exists(self) -> bool:
        return self._path.exists()

    def empty(self) -> bool:
        return not self._data



from __future__ import annotations

import argparse
from typing import TYPE_CHECKING

from pdm.cli.commands.base import BaseCommand
from pdm.cli.commands.venv.backends import BACKENDS
from pdm.cli.options import verbose_option

if TYPE_CHECKING:
    from argparse import ArgumentParser, Namespace

    from pdm.project import Project


class CreateCommand(BaseCommand):
    """Create a virtualenv

    pdm venv create <python> [-other args]
    """

    description = "Create a virtualenv"
    arguments = (verbose_option,)

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "-w",
            "--with",
            dest="backend",
            choices=BACKENDS.keys(),
            help="Specify the backend to create the virtualenv",
        )
        parser.add_argument(
            "-f",
            "--force",
            action="store_true",
            help="Recreate if the virtualenv already exists",
        )
        parser.add_argument("-n", "--name", help="Specify the name of the virtualenv")
        parser.add_argument("--with-pip", action="store_true", help="Install pip with the virtualenv")
        parser.add_argument(
            "python",
            nargs="?",
            help="Specify which python should be used to create the virtualenv",
        )
        parser.add_argument(
            "venv_args",
            nargs=argparse.REMAINDER,
            help="Additional arguments that will be passed to the backend",
        )

    def handle(self, project: Project, options: Namespace) -> None:
        in_project = project.config["venv.in_project"] and not options.name
        backend: str = options.backend or project.config["venv.backend"]
        venv_backend = BACKENDS[backend](project, options.python)
        with project.core.ui.open_spinner(f"Creating virtualenv using [success]{backend}[/]..."):
            path = venv_backend.create(
                options.name,
                options.venv_args,
                options.force,
                in_project,
                prompt=project.config["venv.prompt"],
                with_pip=options.with_pip or project.config["venv.with_pip"],
            )
        project.core.ui.echo(f"Virtualenv [success]{path}[/] is created successfully")



from __future__ import annotations

import email
import email.message
import email.policy
import hashlib
import io
import os
import re
import subprocess
from dataclasses import dataclass
from typing import IO, Any, cast

from pdm.exceptions import PdmUsageError, ProjectError
from pdm.termui import logger

DIST_EXTENSIONS = {
    ".whl": "bdist_wheel",
    ".tar.bz2": "sdist",
    ".tar.gz": "sdist",
    ".zip": "sdist",
}
wheel_file_re = re.compile(
    r"""^(?P<namever>(?P<name>.+?)(-(?P<ver>\d.+?))?)
        ((-(?P<build>\d.*?))?-(?P<pyver>.+?)-(?P<abi>.+?)-(?P<plat>.+?)
        \.whl|\.dist-info)$""",
    re.VERBOSE,
)
UTF8_POLICY = email.policy.EmailPolicy(utf8=True)


def parse_metadata(fp: IO[bytes]) -> email.message.Message:
    """
    Note that this function will close fp. See https://github.com/python/cpython/issues/65562.
    """
    with io.TextIOWrapper(fp, encoding="utf-8", errors="surrogateescape") as file:
        return email.message_from_file(file, policy=UTF8_POLICY)  # type: ignore[arg-type]


@dataclass
class PackageFile:
    """A distribution file for upload.

    XXX: currently only supports sdist and wheel.
    """

    filename: str
    metadata: email.message.Message
    comment: str | None
    py_version: str | None
    filetype: str

    def __post_init__(self) -> None:
        self.base_filename = os.path.basename(self.filename)
        self.gpg_signature: tuple[str, bytes] | None = None

    def get_hashes(self) -> dict[str, str]:
        hashers = {"sha256_digest": hashlib.sha256()}
        try:
            hashers["md5_digest"] = hashlib.md5()
        except ValueError:
            pass
        try:
            hashers["blake2_256_digest"] = hashlib.blake2b(digest_size=256 // 8)  # type: ignore[assignment]
        except (TypeError, ValueError):
            pass
        with open(self.filename, "rb") as f:
            for chunk in iter(lambda: f.read(8192), b""):
                for hasher in hashers.values():
                    hasher.update(chunk)
        return {k: v.hexdigest() for k, v in hashers.items()}

    @classmethod
    def from_filename(cls, filename: str, comment: str | None) -> PackageFile:
        filetype = ""
        for ext, dtype in DIST_EXTENSIONS.items():
            if filename.endswith(ext):
                filetype = dtype
                break
        else:
            raise PdmUsageError(f"Unknown distribution file type: {filename}")
        if filetype == "bdist_wheel":
            metadata = cls.read_metadata_from_wheel(filename)
            match = wheel_file_re.match(os.path.basename(filename))
            if match is None:
                py_ver = "any"
            else:
                py_ver = match.group("pyver")
        elif filename.endswith(".zip"):
            metadata = cls.read_metadata_from_zip(filename)
            py_ver = "source"
        else:
            metadata = cls.read_metadata_from_tar(filename)
            py_ver = "source"
        return cls(filename, metadata, comment, py_ver, filetype)

    @staticmethod
    def read_metadata_from_tar(filename: str) -> email.message.Message:
        import tarfile

        from unearth.preparer import has_leading_dir, split_leading_dir

        if filename.endswith(".gz"):
            mode = "r:gz"
        elif filename.endswith(".bz2"):
            mode = "r:bz2"
        else:
            logger.warning(f"Can't determine the compression mode for {filename}")
            mode = "r:*"
        with tarfile.open(filename, mode) as tar:  # type: ignore[call-overload]
            members = tar.getmembers()
            has_leading = has_leading_dir(m.name for m in members)
            for m in members:
                fn = split_leading_dir(m.name)[1] if has_leading else m.name
                if fn == "PKG-INFO":
                    return parse_metadata(cast(IO[bytes], tar.extractfile(m)))
        raise ProjectError(f"No PKG-INFO found in {filename}")

    @staticmethod
    def read_metadata_from_zip(filename: str) -> email.message.Message:
        import zipfile

        from unearth.preparer import has_leading_dir, split_leading_dir

        with zipfile.ZipFile(filename, allowZip64=True) as zip:
            filenames = zip.namelist()
            has_leading = has_leading_dir(filenames)
            for name in filenames:
                fn = split_leading_dir(name)[1] if has_leading else name
                if fn == "PKG-INFO":
                    return parse_metadata(zip.open(name))
        raise ProjectError(f"No PKG-INFO found in {filename}")

    @staticmethod
    def read_metadata_from_wheel(filename: str) -> email.message.Message:
        import zipfile

        with zipfile.ZipFile(filename, allowZip64=True) as zip:
            for fn in zip.namelist():
                if fn.replace("\\", "/").endswith(".dist-info/METADATA"):
                    return parse_metadata(zip.open(fn))
        raise ProjectError(f"No egg-info is found in {filename}")

    def add_gpg_signature(self, filename: str, signature_name: str) -> None:
        if self.gpg_signature is not None:
            raise PdmUsageError("GPG signature already added")
        with open(filename, "rb") as f:
            self.gpg_signature = (signature_name, f.read())

    def sign(self, identity: str | None) -> None:
        logger.info("Signing %s with gpg", self.base_filename)
        gpg_args = ["gpg", "--detach-sign"]
        if identity is not None:
            gpg_args.extend(["--local-user", identity])
        gpg_args.extend(["-a", self.filename])
        self._run_gpg(gpg_args)
        self.add_gpg_signature(self.filename + ".asc", self.base_filename + ".asc")

    @staticmethod
    def _run_gpg(gpg_args: list[str]) -> None:
        try:
            subprocess.run(gpg_args, check=True)
            return
        except FileNotFoundError:
            logger.warning("gpg executable not available. Attempting fallback to gpg2.")

        gpg_args[0] = "gpg2"
        try:
            subprocess.run(gpg_args, check=True)
        except FileNotFoundError:
            raise PdmUsageError(
                "'gpg' or 'gpg2' executables not available.\n"
                "Try installing one of these or specifying an executable "
                "with the --sign-with flag."
            ) from None

    @property
    def metadata_dict(self) -> dict[str, Any]:
        meta = self.metadata
        data = {
            # identify release
            "name": meta["Name"],
            "version": meta["Version"],
            # file content
            "filetype": self.filetype,
            "pyversion": self.py_version,
            # additional meta-data
            "metadata_version": meta["Metadata-Version"],
            "summary": meta["Summary"],
            "home_page": meta["Home-page"],
            "author": meta["Author"],
            "author_email": meta["Author-email"],
            "maintainer": meta["Maintainer"],
            "maintainer_email": meta["Maintainer-email"],
            "license": meta["License"],
            "description": meta.get_payload(),
            "keywords": meta["Keywords"],
            "platform": meta.get_all("Platform") or (),
            "classifiers": meta.get_all("Classifier") or [],
            "download_url": meta["Download-URL"],
            "supported_platform": meta.get_all("Supported-Platform") or (),
            "comment": self.comment,
            # Metadata 1.2
            "project_urls": meta.get_all("Project-URL") or (),
            "provides_dist": meta.get_all("Provides-Dist") or (),
            "obsoletes_dist": meta.get_all("Obsoletes-Dist") or (),
            "requires_dist": meta.get_all("Requires-Dist") or (),
            "requires_external": meta.get_all("Requires-External") or (),
            "requires_python": meta.get_all("Requires-Python") or (),
            # Metadata 2.1
            "provides_extras": meta.get_all("Provides-Extra") or (),
            "description_content_type": meta.get("Description-Content-Type"),
            # Metadata 2.2
            "dynamic": meta.get_all("Dynamic") or (),
            # Hashes
            **self.get_hashes(),
        }
        if self.gpg_signature is not None:
            data["gpg_signature"] = self.gpg_signature
        return data



from pdm.compat import importlib_metadata, resources_read_text


def read_version() -> str:
    try:
        return importlib_metadata.version(__package__ or "pdm")
    except importlib_metadata.PackageNotFoundError:
        return resources_read_text("pdm", "VERSION").strip()


__version__ = read_version()



from __future__ import annotations

import importlib.resources
import os
import re
import shutil
import subprocess
import tempfile
from copy import deepcopy
from pathlib import Path
from typing import TYPE_CHECKING, Any

from pdm.exceptions import PdmException
from pdm.utils import normalize_name

if TYPE_CHECKING:
    from importlib.resources.abc import Traversable
    from typing import Callable, TypeVar

    ST = TypeVar("ST", Traversable, Path)

TEMPLATE_PACKAGE = "pdm.cli.templates"
BUILTIN_TEMPLATES = ["default", "minimal"]


class ProjectTemplate:
    _path: Path

    def __init__(self, path_or_url: str | None) -> None:
        self.template = path_or_url or "default"

    def __enter__(self) -> ProjectTemplate:
        self._path = Path(tempfile.mkdtemp(suffix="-template", prefix="pdm-"))
        self.prepare_template()
        return self

    def __exit__(self, *args: Any) -> None:
        shutil.rmtree(self._path, ignore_errors=True)

    def generate(self, target_path: Path, metadata: dict[str, Any], overwrite: bool = False) -> None:
        from pdm.compat import tomllib

        def replace_all(path: str, old: str, new: str) -> None:
            with open(path, encoding=encoding) as fp:
                content = fp.read()
            content = re.sub(rf"\b{old}\b", new, content)
            with open(path, "w", encoding=encoding) as fp:
                fp.write(content)

        if metadata.get("project", {}).get("name"):
            try:
                with open(self._path / "pyproject.toml", "rb") as fp:
                    pyproject = tomllib.load(fp)
            except FileNotFoundError:
                raise PdmException("Template pyproject.toml not found") from None
            new_name = metadata["project"]["name"]
            new_import_name = normalize_name(new_name).replace("-", "_")
            try:
                original_name = pyproject["project"]["name"]
            except KeyError:
                raise PdmException("Template pyproject.toml is not PEP-621 compliant") from None
            import_name = normalize_name(original_name).replace("-", "_")
            encoding = "utf-8"
            for root, dirs, filenames in os.walk(self._path):
                for i, d in enumerate(dirs):
                    if d == import_name:
                        os.rename(os.path.join(root, d), os.path.join(root, new_import_name))
                        dirs[i] = new_import_name
                for f in filenames:
                    if f.endswith(".py"):
                        replace_all(os.path.join(root, f), import_name, new_import_name)
                        if f == import_name + ".py":
                            os.rename(os.path.join(root, f), os.path.join(root, new_import_name + ".py"))
                    elif f.endswith((".md", ".rst")):
                        replace_all(os.path.join(root, f), original_name, new_name)
                        replace_all(os.path.join(root, f), import_name, new_import_name)
                    elif Path(root) == self._path and f == "pyproject.toml":
                        replace_all(os.path.join(root, f), original_name, new_name)
                        replace_all(os.path.join(root, f), import_name, new_import_name)

        target_path.mkdir(exist_ok=True, parents=True)
        self.mirror(self._path, target_path, [self._path / "pyproject.toml"], overwrite=overwrite)
        self._generate_pyproject(target_path / "pyproject.toml", metadata)

    def prepare_template(self) -> None:
        if self.template in BUILTIN_TEMPLATES:
            self._prepare_package_template(f"{TEMPLATE_PACKAGE}.{self.template}")
        elif "://" in self.template or self.template.startswith("git@"):
            self._prepare_git_template(self.template)
        elif os.path.exists(self.template):
            self._prepare_local_template(self.template)
        else:  # template name
            template = f"https://github.com/pdm-project/template-{self.template}"
            self._prepare_git_template(template)

    @staticmethod
    def mirror(
        src: ST,
        dst: Path,
        skip: list[ST] | None = None,
        copyfunc: Callable[[ST, Path], Any] = shutil.copyfile,  # type: ignore[assignment]
        *,
        overwrite: bool = False,
    ) -> None:
        if skip and src in skip:
            return
        if src.is_dir():
            dst.mkdir(exist_ok=True)
            for child in src.iterdir():
                ProjectTemplate.mirror(child, dst / child.name, skip, copyfunc)
        elif src.name.endswith(".pyc"):
            return
        elif overwrite or not dst.exists():
            copyfunc(src, dst)

    @staticmethod
    def _copy_package_file(src: Traversable, dst: Path) -> Path:
        with importlib.resources.as_file(src) as f:
            return shutil.copyfile(f, dst)

    def _generate_pyproject(self, path: Path, metadata: dict[str, Any]) -> None:
        import tomlkit

        from pdm.cli.utils import merge_dictionary

        try:
            with open(path, encoding="utf-8") as fp:
                content = tomlkit.load(fp)
        except FileNotFoundError:
            content = tomlkit.document()
        try:
            with open(self._path / "pyproject.toml", encoding="utf-8") as fp:
                template_content = tomlkit.load(fp)
        except FileNotFoundError:
            template_content = tomlkit.document()

        # repeated calls to merge_dictionary could extend elements in the template build system, get a deep copy now
        template_build_system = deepcopy(template_content.get("build-system", {}))
        merge_dictionary(content, template_content)
        if "version" in content.get("project", {}).get("dynamic", []):
            metadata["project"].pop("version", None)
        merge_dictionary(content, metadata)
        if "build-system" in metadata:
            build_system = metadata["build-system"]
            if template_build_system.get("build-backend") == build_system["build-backend"]:
                # only merge the build system when the selected build backend matches that of the template
                merge_dictionary(build_system, template_build_system)
            content["build-system"] = build_system
        else:
            content.pop("build-system", None)
        with open(path, "w", encoding="utf-8") as fp:
            fp.write(tomlkit.dumps(content))

    def _prepare_package_template(self, import_name: str) -> None:
        files = importlib.resources.files(import_name)

        self.mirror(files, self._path, skip=[files / "__init__.py"], copyfunc=self._copy_package_file)

    def _prepare_git_template(self, url: str) -> None:
        left, amp, right = url.rpartition("@")
        if left != "git" and amp:
            extra_args = [f"--branch={right}"]
            url = left
        else:
            extra_args = []
        git_command = ["git", "clone", "--recursive", "--depth=1", *extra_args, url, self._path.as_posix()]
        result = subprocess.run(git_command, capture_output=True, text=True)
        if result.returncode != 0:
            raise PdmException(f"Failed to clone template from git repository {url}: {result.stderr}")
        shutil.rmtree(self._path / ".git", ignore_errors=True)

    def _prepare_local_template(self, path: str) -> None:
        src = Path(path)

        self.mirror(src, self._path, skip=[src / ".git", src / ".svn", src / ".hg"])



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pdm.models.candidates import Candidate


class PdmException(Exception):
    pass


class ResolutionError(PdmException):
    pass


class PdmArgumentError(PdmException):
    pass


class PdmUsageError(PdmException):
    pass


class RequirementError(PdmUsageError, ValueError):
    pass


class PublishError(PdmUsageError):
    pass


class InvalidPyVersion(PdmUsageError, ValueError):
    pass


class CandidateNotFound(PdmException):
    pass


class CandidateInfoNotFound(PdmException):
    def __init__(self, candidate: Candidate) -> None:
        message = f"No metadata information is available for [success]{candidate!s}[/]."
        self.candidate = candidate
        super().__init__(message)


class PDMWarning(Warning):
    pass


class PackageWarning(PDMWarning):
    pass


class PDMDeprecationWarning(PDMWarning, DeprecationWarning):
    pass


warnings.simplefilter("default", category=PDMDeprecationWarning)


class ProjectError(PdmUsageError):
    pass


class InstallationError(PdmException):
    pass


class UninstallError(PdmException):
    pass


class NoConfigError(PdmUsageError, KeyError):
    def __str__(self) -> str:
        return f"No such config key: {self.args[0]!r}"


class NoPythonVersion(PdmUsageError):
    pass


class BuildError(PdmException, RuntimeError):
    pass



from __future__ import annotations

from functools import cached_property
from typing import TYPE_CHECKING, Iterable

import tomlkit

from pdm.project.lockfile.base import (
    FLAG_CROSS_PLATFORM,
    FLAG_DIRECT_MINIMAL_VERSIONS,
    FLAG_INHERIT_METADATA,
    FLAG_STATIC_URLS,
    Compatibility,
    Lockfile,
)
from pdm.utils import parse_version

if TYPE_CHECKING:
    from pdm.models.repositories import LockedRepository


class PDMLock(Lockfile):
    SUPPORTED_FLAGS = frozenset(
        (FLAG_STATIC_URLS, FLAG_CROSS_PLATFORM, FLAG_DIRECT_MINIMAL_VERSIONS, FLAG_INHERIT_METADATA)
    )
    spec_version = parse_version("4.5.0")

    @property
    def hash(self) -> tuple[str, str]:
        content_hash = self._data.get("metadata", {}).get("content_hash", "")
        return content_hash.split(":", 1)

    @property
    def file_version(self) -> str:
        return self._data.get("metadata", {}).get("lock_version", "")

    @property
    def groups(self) -> list[str] | None:
        return self._data.get("metadata", {}).get("groups")

    @cached_property
    def default_strategies(self) -> set[str]:
        return {FLAG_INHERIT_METADATA}

    @property
    def strategy(self) -> set[str]:
        metadata = self._data.get("metadata", {})
        if not metadata:
            return self.default_strategies.copy()
        result: set[str] = set(metadata.get("strategy", {FLAG_CROSS_PLATFORM}))
        # Compatibility with old lockfiles
        if not metadata.get(FLAG_CROSS_PLATFORM, True):
            result.discard(FLAG_CROSS_PLATFORM)
        if metadata.get(FLAG_STATIC_URLS, False):
            result.add(FLAG_STATIC_URLS)
        return result & self.SUPPORTED_FLAGS

    def update_hash(self, hash_value: str, algo: str = "sha256") -> None:
        self._data.setdefault("metadata", {})["content_hash"] = f"{algo}:{hash_value}"

    def compatibility(self) -> Compatibility:
        """We use a three-part versioning scheme for lockfiles:
        The first digit represents backward compatibility and the second digit represents forward compatibility.
        """
        if not self.exists():
            return Compatibility.SAME
        if not self.file_version:
            return Compatibility.NONE
        lockfile_version = parse_version(self.file_version)
        if lockfile_version == self.spec_version:
            return Compatibility.SAME
        if lockfile_version.major != self.spec_version.major or lockfile_version.minor > self.spec_version.minor:
            return Compatibility.NONE
        if lockfile_version.minor < self.spec_version.minor:
            return Compatibility.BACKWARD
        return Compatibility.BACKWARD if lockfile_version.micro < self.spec_version.micro else Compatibility.FORWARD

    def format_lockfile(self, repository: LockedRepository, groups: Iterable[str] | None, strategy: set[str]) -> None:
        """Format lock file from a dict of resolved candidates, a mapping of dependencies
        and a collection of package summaries.
        """
        from pdm.formats.base import make_array, make_inline_table

        def _group_sort_key(group: str) -> tuple[bool, str]:
            return group != "default", group

        project = repository.environment.project
        packages = tomlkit.aot()
        for entry in sorted(repository.packages.values(), key=lambda x: x.candidate.identify()):
            base = tomlkit.table()
            base.update(entry.candidate.as_lockfile_entry(project.root))
            base.add("summary", entry.summary or "")
            if FLAG_INHERIT_METADATA in strategy:
                base.add("groups", sorted(entry.candidate.req.groups, key=_group_sort_key))
                if entry.candidate.req.marker is not None:
                    base.add("marker", str(entry.candidate.req.marker))
            if entry.dependencies:
                base.add("dependencies", make_array(sorted(entry.dependencies), True))
            if hashes := entry.candidate.hashes:
                collected = {}
                for item in hashes:
                    if FLAG_STATIC_URLS in strategy:
                        row = {"url": item["url"], "hash": item["hash"]}
                    else:
                        row = {"file": item["file"], "hash": item["hash"]}
                    inline = make_inline_table(row)
                    # deduplicate and sort
                    collected[tuple(row.values())] = inline
                if collected:
                    base.add("files", make_array([collected[k] for k in sorted(collected)], True))
            packages.append(base)
        doc = tomlkit.document()
        metadata = tomlkit.table()
        if groups is None:
            groups = list(project.iter_groups())
        metadata.update(
            {
                "groups": sorted(groups, key=_group_sort_key),
                "strategy": sorted(strategy),
                "targets": [t.as_dict() for t in repository.targets],
                "lock_version": str(self.spec_version),
            }
        )
        metadata.pop(FLAG_STATIC_URLS, None)
        metadata.pop(FLAG_CROSS_PLATFORM, None)
        doc.add("metadata", metadata)
        doc.add("package", packages)
        self.set_data(doc)



from __future__ import annotations

import contextlib
import datetime
import hashlib
import inspect
import json
import os
import sys
import textwrap
from typing import TYPE_CHECKING, Collection, Iterable, cast

from resolvelib.resolvers import ResolutionImpossible, ResolutionTooDeep

from pdm import termui
from pdm.cli.filters import GroupSelection
from pdm.cli.hooks import HookManager
from pdm.cli.utils import (
    check_project_file,
    find_importable_files,
    format_resolution_impossible,
    get_pep582_path,
    set_env_in_reg,
)
from pdm.environments import BareEnvironment
from pdm.exceptions import PdmException, PdmUsageError, ProjectError, ResolutionError
from pdm.models.candidates import Candidate
from pdm.models.markers import EnvSpec
from pdm.models.repositories import LockedRepository, Package
from pdm.project import Project
from pdm.project.lockfile import FLAG_CROSS_PLATFORM, FLAG_INHERIT_METADATA, FLAG_STATIC_URLS
from pdm.resolver.reporters import RichLockReporter
from pdm.termui import logger
from pdm.utils import deprecation_warning

if TYPE_CHECKING:
    from pdm.models.requirements import Requirement


def do_lock(
    project: Project,
    strategy: str = "all",
    tracked_names: Iterable[str] | None = None,
    requirements: list[Requirement] | None = None,
    dry_run: bool = False,
    refresh: bool = False,
    groups: list[str] | None = None,
    strategy_change: list[str] | None = None,
    hooks: HookManager | None = None,
    env_spec: EnvSpec | None = None,
    append: bool = False,
) -> dict[str, list[Candidate]]:
    """Performs the locking process and update lockfile."""
    hooks = hooks or HookManager(project)
    check_project_file(project)
    lock_strategy = project.lockfile.apply_strategy_change(strategy_change or [])
    if FLAG_CROSS_PLATFORM in lock_strategy:  # pragma: no cover
        project.core.ui.deprecated(
            "`cross_platform` strategy is deprecated in favor of the new lock targets.\n"
            "See docs: http://pdm-project.org/en/latest/usage/lock-targets/"
        )
    if strategy_change and append:
        raise PdmUsageError("Not allowed to change lock strategy when --append is used.")
    locked_repo = project.get_locked_repository()
    candidates = [entry.candidate for entry in locked_repo.packages.values()]
    if refresh or (env_spec is not None and not append):
        # Refetch hashes if --platform/--python/--implementation/--refresh is used
        for c in candidates:
            c.hashes.clear()
    if refresh:
        if env_spec is not None:
            raise PdmUsageError("Cannot pass --python/--platform/--implementation with --refresh")
        repo = project.get_repository()
        with project.core.ui.open_spinner("Re-calculating hashes..."):
            with project.core.ui.logging("lock"):
                repo.fetch_hashes(candidates)
            project.lockfile.format_lockfile(locked_repo, groups=project.lockfile.groups, strategy=lock_strategy)
        project.write_lockfile()
        return locked_repo.all_candidates

    if groups is None:
        groups = list(project.iter_groups())
    if not requirements:
        all_deps = project._resolve_dependencies(groups)
        requirements = [r for group in groups for r in project.get_dependencies(group, all_deps)]
    ui = project.core.ui
    supports_env_spec = "env_spec" in inspect.signature(project.get_provider).parameters
    # The repository to store the lock result
    if locked_repo.packages:
        result_repo = LockedRepository({}, sources=project.sources, environment=project.environment)
    else:
        # Use the same repository if the lock is empty.
        result_repo = locked_repo
    if not supports_env_spec:  # pragma: no cover
        ui.warn("Lock targets are not supported by the current provider")

    if append:
        if env_spec is None:
            raise PdmUsageError("Cannot use `--append` without --python/--platform/--implementation")
        if env_spec in locked_repo.targets:
            ui.echo(f"{termui.Emoji.LOCK} Lock target {env_spec} already exists, skip locking.")
            return locked_repo.all_candidates
        targets = [env_spec]
        result_repo = locked_repo  # Append to the same lockfile
    else:
        targets = [env_spec] if env_spec else (locked_repo.targets[:] or project.lock_targets)
    # Restrict the target python to within the project's requires-python
    global_requires_python = project.environment.python_requires
    for i, target in enumerate(targets):
        if (merged := global_requires_python & target.requires_python).is_empty():
            raise PdmUsageError(
                f"The target requires Python {target.requires_python} which is not compatible with "
                f"the project's requires-python {global_requires_python}"
            )
        targets[i] = target.replace(requires_python=merged)
    resolve_max_rounds = int(project.config["strategy.resolve_max_rounds"])
    hooks.try_emit("pre_lock", requirements=requirements, dry_run=dry_run)
    with ui.logging("lock"):
        for req in list(requirements):
            if req.key == "python":
                requirements.remove(req)
                logger.warning("The 'python' requirement is not necessary and will be ignored.")
        # The context managers are nested to ensure the spinner is stopped before
        # any message is thrown to the output.
        resolver_class = project.get_resolver()
        with RichLockReporter(requirements, ui) as reporter:
            collected_groups: set[str] = set()
            try:
                for target in targets:
                    resolver = resolver_class(
                        environment=project.environment,
                        requirements=[r for r in requirements if not r.marker or r.marker.matches(target)],
                        update_strategy=strategy,
                        strategies=lock_strategy,
                        target=target,
                        tracked_names=list(tracked_names or ()),
                        locked_repository=locked_repo,
                        reporter=reporter,
                    )
                    reporter.update(f"Resolve for environment {target}")
                    resolved, new_groups = resolver.resolve()
                    collected_groups.update(new_groups)
                    locked_repo.merge_result(target, resolved)
                    if result_repo is not locked_repo:
                        result_repo.merge_result(target, resolved)
            except ResolutionTooDeep:
                reporter.update(f"{termui.Emoji.LOCK} Lock failed.", info="", completed=1)
                ui.echo(
                    "The dependency resolution exceeds the maximum loop depth of "
                    f"{resolve_max_rounds}, there may be some circular dependencies "
                    "in your project. Try to solve them or increase the "
                    f"[success]`strategy.resolve_max_rounds`[/] config.",
                    err=True,
                )
                raise ResolutionError(
                    f"The dependency resolution exceeds the maximum loop depth of {resolve_max_rounds}"
                ) from None
            except ResolutionImpossible as err:
                reporter.update(f"{termui.Emoji.LOCK} Lock failed.", info="", completed=1)
                ui.error(format_resolution_impossible(err))
                raise ResolutionError("Unable to find a resolution") from None
            else:
                groups = list(set(groups) | collected_groups)
                if project.enable_write_lockfile:
                    reporter.update(f"{termui.Emoji.LOCK} Lock successful.", info="", completed=1)
                project.lockfile.format_lockfile(result_repo, groups=groups, strategy=lock_strategy)
                project.write_lockfile(write=not dry_run)
                hooks.try_emit("post_lock", resolution=result_repo.all_candidates, dry_run=dry_run)

    return result_repo.all_candidates


def resolve_from_lockfile(
    project: Project,
    requirements: Iterable[Requirement],
    groups: Collection[str] | None = None,
    env_spec: EnvSpec | None = None,
) -> Iterable[Package]:
    from dep_logic.tags import EnvCompatibility

    from pdm.resolver.resolvelib import RLResolver

    ui = project.core.ui

    if env_spec is None:
        # Resolve for the current environment by default
        env_spec = project.environment.spec
    reqs = [req for req in requirements if not req.marker or req.marker.matches(env_spec)]
    with ui.open_spinner("Resolving packages from lockfile..."):
        locked_repo = project.get_locked_repository(env_spec)
        lock_targets = locked_repo.targets
        if env_spec not in lock_targets:
            compatibilities = [target.compare(env_spec) for target in lock_targets]
            if not any(compat == EnvCompatibility.LOWER_OR_EQUAL for compat in compatibilities):
                loose_compatible_target = next(
                    (
                        target
                        for (target, compat) in zip(lock_targets, compatibilities)
                        if compat == EnvCompatibility.HIGHER
                    ),
                    None,
                )
                if loose_compatible_target is not None:
                    ui.warn(f"Found lock target {loose_compatible_target}, installing for env {env_spec}")
                elif not lock_targets:  # pragma: no cover
                    ui.warn("Missing lock targets or environment field in the lock file, installing it anyway.")
                else:
                    errors = [f"None of the lock targets matches the current env {env_spec}:"] + [
                        f" - {target}" for target in lock_targets
                    ]
                    ui.error("\n".join(errors))
                    raise PdmException("No compatible lock target found")

        with ui.logging("install-resolve"):
            strategies = project.lockfile.strategy.copy()
            if FLAG_INHERIT_METADATA in strategies and groups is not None and not project.config["use_uv"]:
                return locked_repo.evaluate_candidates(groups)
            strategies.add(FLAG_STATIC_URLS)
            resolver = project.get_resolver()(
                environment=project.environment,
                requirements=reqs,
                update_strategy="reuse",
                strategies=strategies,
                target=env_spec,
                tracked_names=(),
                locked_repository=locked_repo,
            )
            if isinstance(resolver, RLResolver):  # resolve from lock file
                resolver.provider.repository = locked_repo
            try:
                return resolver.resolve().packages
            except ResolutionImpossible as e:
                logger.exception("Broken lockfile")
                raise PdmException(
                    "Resolving from lockfile failed. You may fix the lockfile by `pdm lock --update-reuse` and retry."
                ) from e


def resolve_candidates_from_lockfile(
    project: Project,
    requirements: Iterable[Requirement],
    cross_platform: bool | None = None,
    groups: Collection[str] | None = None,
    env_spec: EnvSpec | None = None,
) -> dict[str, Candidate]:
    if cross_platform is not None:  # pragma: no cover
        deprecation_warning("cross_platform argument is deprecated", stacklevel=2)
    packages = resolve_from_lockfile(project, requirements, groups, env_spec)
    return {p.candidate.identify(): p.candidate for p in packages}


def check_lockfile(project: Project, raise_not_exist: bool = True) -> str | None:
    """Check if the lock file exists and is up to date. Return the lock strategy."""
    from pdm.project.lockfile.base import Compatibility

    if not project.lockfile.exists():
        if raise_not_exist:
            raise ProjectError("Lockfile does not exist, nothing to install")
        project.core.ui.warn("Lockfile does not exist")
        return "all"
    compat = project.lockfile.compatibility()
    if compat == Compatibility.NONE:
        project.core.ui.warn("Lockfile is not compatible with PDM")
        return "reuse"
    elif compat == Compatibility.BACKWARD:
        project.core.ui.warn("Lockfile is generated on an older version of PDM")
    elif compat == Compatibility.FORWARD:
        project.core.ui.warn("Lockfile is generated on a newer version of PDM")
    if not project.is_lockfile_hash_match():
        project.core.ui.warn("Lockfile hash doesn't match pyproject.toml, packages may be outdated")
        return "reuse"
    return None


def do_sync(
    project: Project,
    *,
    selection: GroupSelection,
    dry_run: bool = False,
    clean: bool = False,
    quiet: bool = False,
    requirements: list[Requirement] | None = None,
    tracked_names: Collection[str] | None = None,
    no_editable: bool | Collection[str] = False,
    no_self: bool = False,
    reinstall: bool = False,
    only_keep: bool = False,
    fail_fast: bool = False,
    hooks: HookManager | None = None,
) -> None:
    """Synchronize project"""
    hooks = hooks or HookManager(project)
    if requirements is None:
        requirements = []
        selection.validate()
        all_deps = project._resolve_dependencies(list(selection))
        for group in selection:
            requirements.extend(project.get_dependencies(group, all_deps))
    packages = list(resolve_from_lockfile(project, requirements, groups=list(selection)))
    if tracked_names and dry_run:
        packages = [p for p in packages if p.candidate.identify() in tracked_names]
    synchronizer = project.get_synchronizer(quiet=quiet)(
        project.environment,
        clean=clean,
        dry_run=dry_run,
        no_editable=no_editable,
        install_self=not no_self and project.is_distribution,
        reinstall=reinstall,
        only_keep=only_keep,
        fail_fast=fail_fast,
        packages=packages,
    )
    with project.core.ui.logging("install"):
        hooks.try_emit("pre_install", packages=packages, dry_run=dry_run)
        synchronizer.synchronize()
        hooks.try_emit("post_install", packages=packages, dry_run=dry_run)


def ask_for_import(project: Project) -> None:
    """Show possible importable files and ask user to decide"""
    from pdm.cli.commands.import_cmd import Command as ImportCommand

    importable_files = list(find_importable_files(project))
    if not importable_files:
        return
    project.core.ui.echo("Found following files from other formats that you may import:", style="primary")
    for i, (key, filepath) in enumerate(importable_files):
        project.core.ui.echo(f"{i}. [success]{filepath.as_posix()}[/] ({key})")
    project.core.ui.echo(f"{len(importable_files)}. [warning]don't do anything, I will import later.[/]")
    choice = termui.ask(
        "Please select",
        prompt_type=int,
        choices=[str(i) for i in range(len(importable_files) + 1)],
        show_choices=False,
    )
    if int(choice) == len(importable_files):
        return
    key, filepath = importable_files[int(choice)]
    ImportCommand.do_import(project, str(filepath), key, reset_backend=False)


def print_pep582_command(project: Project, shell: str = "AUTO") -> None:
    """Print the export PYTHONPATH line to be evaluated by the shell."""
    import shellingham

    pep582_path = get_pep582_path(project)
    ui = project.core.ui

    if os.name == "nt":
        try:
            set_env_in_reg("PYTHONPATH", pep582_path)
        except PermissionError:
            ui.error("Permission denied, please run the terminal as administrator.")
        ui.info("The environment variable has been saved, please restart the session to take effect.")
        return
    lib_path = pep582_path.replace("'", "\\'")
    if shell == "AUTO":
        shell = shellingham.detect_shell()[0]
    shell = shell.lower()
    if shell in ("zsh", "bash", "sh", "dash"):
        result = textwrap.dedent(
            f"""
            if [ -n "$PYTHONPATH" ]; then
                export PYTHONPATH='{lib_path}':$PYTHONPATH
            else
                export PYTHONPATH='{lib_path}'
            fi
            """
        ).strip()
    elif shell == "fish":
        result = f"set -x PYTHONPATH '{lib_path}' $PYTHONPATH"
    elif shell in ("tcsh", "csh"):
        result = textwrap.dedent(
            f"""
            if ( $?PYTHONPATH ) then
                if ( "$PYTHONPATH" != "" ) then
                    setenv PYTHONPATH '{lib_path}':$PYTHONPATH
                else
                    setenv PYTHONPATH '{lib_path}'
                endif
            else
                setenv PYTHONPATH '{lib_path}'
            endif
            """
        ).strip()
    else:
        raise PdmUsageError(f"Unsupported shell: {shell}, please specify another shell via `--pep582 <SHELL>`")
    ui.echo(result)


def get_latest_pdm_version_from_pypi(project: Project, prereleases: bool = False) -> str | None:
    """Get the latest version of PDM from PyPI."""
    environment = BareEnvironment(project)
    with environment.get_finder([project.default_source]) as finder:
        candidate = finder.find_best_match("pdm", allow_prereleases=prereleases).best
    return cast(str, candidate.version) if candidate else None


def get_latest_version(project: Project, expire_after: int = 7 * 24 * 3600) -> str | None:  # pragma: no cover
    """Get the latest version of PDM from PyPI, cache for 7 days"""
    cache_key = hashlib.sha224(sys.executable.encode()).hexdigest()
    cache_file = project.cache("self-check") / cache_key
    state = {}
    with contextlib.suppress(OSError):
        state = json.loads(cache_file.read_text())
    current_time = datetime.datetime.now(datetime.timezone.utc).timestamp()
    if (last_check := state.get("last-check")) and current_time - last_check < expire_after:
        return cast(str, state["latest-version"])
    try:
        latest_version = get_latest_pdm_version_from_pypi(project)
    except Exception as e:
        project.core.ui.warn(f"Failed to get latest version: {e}", verbosity=termui.Verbosity.NORMAL)
        latest_version = None
    if latest_version is None:
        return None
    state.update({"latest-version": latest_version, "last-check": current_time})
    with contextlib.suppress(OSError):
        cache_file.write_text(json.dumps(state))
    return latest_version


def check_update(project: Project) -> None:  # pragma: no cover
    """Check if there is a new version of PDM available"""
    from pdm.cli.utils import is_homebrew_installation, is_pipx_installation, is_scoop_installation
    from pdm.utils import parse_version

    if project.core.ui.verbosity < termui.Verbosity.NORMAL:
        return

    this_version = project.core.version
    latest_version = get_latest_version(project)
    if latest_version is None or parse_version(this_version) >= parse_version(latest_version):
        return
    disable_command = "pdm config check_update false"

    is_prerelease = parse_version(latest_version).is_prerelease

    if is_pipx_installation():
        install_command = f"pipx upgrade {'--pip-args=--pre ' if is_prerelease else ''}pdm"
    elif is_homebrew_installation():
        install_command = "brew upgrade pdm"
    elif is_scoop_installation():
        install_command = "scoop update pdm"
    else:
        install_command = "pdm self update" + (" --pre" if is_prerelease else "")
        if os.name == "nt":
            # On Windows, the executable can't replace itself, we add the python prefix to the command
            # A bit ugly but it works
            install_command = f"{sys.executable} -m {install_command}"

    message = [
        f"PDM [primary]{this_version}[/]",
        f" is installed, while [primary]{latest_version}[/]",
        " is available.\n",
        f"Please run [req]`{install_command}`[/]",
        " to upgrade.\n",
        f"Run [req]`{disable_command}`[/]",
        " to disable the check.",
    ]
    project.core.ui.info("".join(message))



from __future__ import annotations

import os
import shutil
import sys
import tempfile
from argparse import ArgumentParser
from pathlib import Path
from typing import TYPE_CHECKING, cast

from pdm.cli.commands.base import BaseCommand
from pdm.cli.options import verbose_option
from pdm.environments import BareEnvironment
from pdm.exceptions import InstallationError, PdmArgumentError
from pdm.models.python import PythonInfo
from pdm.termui import Verbosity
from pdm.utils import get_all_installable_python_versions

if TYPE_CHECKING:
    from argparse import ArgumentParser, Namespace, _SubParsersAction
    from typing import Any

    from pdm.project.core import Project


class Command(BaseCommand):
    """Manage installed Python interpreters"""

    arguments = ()

    def add_arguments(self, parser: ArgumentParser) -> None:
        self.parser = parser
        subparsers = parser.add_subparsers(title="commands", metavar="")
        ListCommand.register_to(subparsers, name="list")
        RemoveCommand.register_to(subparsers, name="remove")
        InstallCommand.register_to(subparsers, name="install")
        LinkCommand.register_to(subparsers, name="link")
        FindCommand.register_to(subparsers, name="find")

    @classmethod
    def register_to(cls, subparsers: _SubParsersAction, name: str | None = None, **kwargs: Any) -> None:
        return super().register_to(subparsers, name, aliases=["py"], **kwargs)

    def handle(self, project: Project, options: Namespace) -> None:
        self.parser.print_help()


class ListCommand(BaseCommand):
    """List all Python interpreters installed with PDM"""

    arguments = (verbose_option,)

    def handle(self, project: Project, options: Namespace) -> None:
        from findpython.providers.rye import RyeProvider

        ui = project.core.ui
        provider = RyeProvider(root=Path(project.config["python.install_root"]).expanduser())
        for version in provider.find_pythons():
            ui.echo(f"[success]{version}[/] ({version.executable})")


class RemoveCommand(BaseCommand):
    """Remove a Python interpreter installed with PDM"""

    arguments = (verbose_option,)

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("version", help="The Python version to remove. E.g. cpython@3.10.3")

    def handle(self, project: Project, options: Namespace) -> None:
        ui = project.core.ui
        root = Path(project.config["python.install_root"]).expanduser()
        if not root.exists():
            ui.error(f"No Python interpreter found for {options.version!r}")
            sys.exit(1)
        version = str(options.version)
        if root.joinpath(version).exists():
            version_dir = root.joinpath(version)
        else:
            version = options.version.lower()
            if "@" not in version:  # pragma: no cover
                version = f"cpython@{version}"
            version_dir = root.joinpath(version)
            if not version_dir.exists():
                ui.error(f"No Python interpreter found for {options.version!r}")
                ui.echo("Installed Pythons:", err=True)
                for child in root.iterdir():
                    if not child.name.startswith("."):
                        ui.echo(f"  {child.name}", err=True)
                sys.exit(1)
        if version_dir.is_symlink():
            version_dir.unlink()
        else:
            shutil.rmtree(version_dir, ignore_errors=True)
        ui.echo(f"[success]Removed installed[/] {options.version}", verbosity=Verbosity.NORMAL)


class InstallCommand(BaseCommand):
    """Install a Python interpreter with PDM"""

    arguments = (verbose_option,)

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument(
            "version",
            help="The Python version to install (e.g. cpython@3.10.3). If left empty, "
            "highest cPython version that matches this platform/arch is installed. "
            "If pyproject.toml with requires-python is available, this is considered as well.",
            nargs="?",
        )
        parser.add_argument("--list", "-l", action="store_true", help="List all available Python versions")
        parser.add_argument(
            "--min",
            action="store_true",
            help="Use minimum instead of highest version for installation if `version` is left empty",
        )

    def handle(self, project: Project, options: Namespace) -> None:
        if options.list:
            for version in get_all_installable_python_versions(build_dir=False):
                project.core.ui.echo(str(version))
            return
        version = options.version
        if version is None:
            match = project.get_best_matching_cpython_version(options.min)
            if match is not None:
                version = str(match)

        if version is None:
            raise PdmArgumentError("Please specify a Python version to be installed. E.g. cpython@3.10.3")

        self.install_python(project, version)

    @staticmethod
    def install_python(project: Project, request: str) -> PythonInfo:
        from pbs_installer import download, get_download_link, install_file
        from pbs_installer._install import THIS_ARCH

        from pdm.misc import sysconfig_patcher
        from pdm.termui import logger

        ui = project.core.ui
        root = Path(project.config["python.install_root"]).expanduser()

        implementation, _, version = request.rpartition("@")
        implementation = implementation.lower() or "cpython"
        version, _, arch = version.partition("-")
        arch = "x86" if arch == "32" else (arch or THIS_ARCH)

        ver, python_file = get_download_link(version, implementation=implementation, arch=arch, build_dir=False)
        ver_str = str(ver)
        spinner_msg = f"Downloading [success]{ver_str}[/]"
        if ui.verbosity >= Verbosity.DETAIL:
            download_url = python_file[0] if isinstance(python_file, (tuple, list)) else python_file
            spinner_msg += f" {download_url}"

        with ui.open_spinner(spinner_msg) as spinner:
            destination = root / ver_str
            logger.debug("Installing %s to %s", ver_str, destination)
            env = BareEnvironment(project)
            install_root = destination
            if install_root.joinpath("install").exists():
                install_root = install_root.joinpath("install")
            interpreter = install_root / "bin" / "python3" if sys.platform != "win32" else destination / "python.exe"
            if not destination.exists() or not interpreter.exists():
                shutil.rmtree(destination, ignore_errors=True)
                destination.mkdir(parents=True, exist_ok=True)
                with tempfile.NamedTemporaryFile() as tf:
                    tf.close()
                    original_filename = download(python_file, tf.name, env.session)
                    spinner.update(f"Installing [success]{ver_str}[/]")
                    try:
                        install_file(tf.name, destination, original_filename)
                    except ModuleNotFoundError as e:
                        if "zstandard is required" in str(e):
                            raise InstallationError(
                                "zstandard is required to install this Python version. "
                                "Please install it with `pdm self add zstandard`."
                            ) from None
        if destination.joinpath("install").exists():
            install_root = destination.joinpath("install")
            interpreter = install_root / "bin" / "python3" if sys.platform != "win32" else install_root / "python.exe"
        if not interpreter.exists():
            raise InstallationError("Installation failed, please try again.")
        # Patch sysconfig and pkgconfig files with correct prefixes
        sysconfig_patcher.patch(install_root)
        python_info = PythonInfo.from_path(interpreter)
        ui.echo(
            f"[success]Successfully installed[/] {python_info.implementation}@{python_info.version}",
            verbosity=Verbosity.NORMAL,
        )
        ui.echo(f"[info]Version:[/] {python_info.version}", verbosity=Verbosity.NORMAL)
        ui.echo(f"[info]Executable:[/] {python_info.path}", verbosity=Verbosity.NORMAL)
        return python_info


class LinkCommand(BaseCommand):
    """Link an external Python interpreter to PDM"""

    arguments = (verbose_option,)

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("interpreter", help="The path to the Python interpreter to link")
        parser.add_argument("--name", help="The name of the link")

    def handle(self, project: Project, options: Namespace) -> None:
        python_info = PythonInfo.from_path(options.interpreter)
        if not python_info.valid:
            raise PdmArgumentError("Invalid Python interpreter")
        if options.name is None:
            link_name = f"{python_info.implementation}@{python_info.identifier}"
        else:
            link_name = cast(str, options.name)
        link_path = Path(project.config["python.install_root"]).expanduser() / link_name
        if link_path.exists():
            raise PdmArgumentError(f"Link {link_name} already exists")
        exe_dir = python_info.path.parent
        if exe_dir.name in ("Scripts", "bin"):
            exe_dir = exe_dir.parent
        link_path.parent.mkdir(parents=True, exist_ok=True)
        link_path.symlink_to(exe_dir)
        project.core.ui.echo(f"[success]Successfully linked {link_name} to {exe_dir}[/]")


class FindCommand(BaseCommand):
    """Search for a Python interpreter"""

    arguments = (verbose_option,)

    def add_arguments(self, parser: ArgumentParser) -> None:
        parser.add_argument("--managed", action="store_true", help="Only find interpreters managed by PDM")
        parser.add_argument("request", help="The Python version to find. E.g. 3.12, cpython@3.13")

    def handle(self, project: Project, options: Namespace) -> None:
        from findpython import Finder

        if options.managed:
            old_rye_root = os.getenv("RYE_PY_ROOT")
            os.environ["RYE_PY_ROOT"] = os.path.expanduser(project.config["python.install_root"])
            try:
                finder = Finder(resolve_symlinks=True, selected_providers=["rye"])
            finally:
                if old_rye_root:  # pragma: no cover
                    os.environ["RYE_PY_ROOT"] = old_rye_root
                else:
                    del os.environ["RYE_PY_ROOT"]
        else:
            finder = project._get_python_finder()

        python_version = finder.find(options.request)
        location = (
            "managed installations"
            if options.managed
            else "virtual environments, managed installations, or search paths"
        )
        if python_version is None:
            project.core.ui.error(f"Python interpreter {options.request!r} not found in {location}")
            raise SystemExit(1)
        print(python_version.executable)



from __future__ import annotations

import argparse
from collections import defaultdict
from typing import TYPE_CHECKING

from pdm.cli.commands.base import BaseCommand
from pdm.cli.filters import GroupSelection
from pdm.cli.hooks import HookManager
from pdm.cli.options import (
    frozen_lockfile_option,
    groups_group,
    install_group,
    lockfile_option,
    override_option,
    prerelease_option,
    save_strategy_group,
    skip_option,
    unconstrained_option,
    update_strategy_group,
    venv_option,
)
from pdm.exceptions import PdmUsageError, ProjectError

if TYPE_CHECKING:
    from typing import Collection

    from pdm.models.requirements import Requirement
    from pdm.project import Project


class Command(BaseCommand):
    """Update package(s) in pyproject.toml"""

    arguments = (
        *BaseCommand.arguments,
        groups_group,
        install_group,
        lockfile_option,
        frozen_lockfile_option,
        save_strategy_group,
        override_option,
        update_strategy_group,
        prerelease_option,
        unconstrained_option,
        skip_option,
        venv_option,
    )

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "-t",
            "--top",
            action="store_true",
            help="Only update those listed in pyproject.toml",
        )
        parser.add_argument(
            "--dry-run",
            "--outdated",
            action="store_true",
            dest="dry_run",
            help="Show the difference only without modifying the lockfile content",
        )
        parser.add_argument(
            "--no-sync",
            dest="sync",
            default=True,
            action="store_false",
            help="Only update lock file but do not sync packages",
        )
        parser.add_argument("packages", nargs="*", help="If packages are given, only update them")
        parser.set_defaults(dev=None)

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        self.do_update(
            project,
            selection=GroupSelection.from_options(project, options),
            save=options.save_strategy or project.config["strategy.save"],
            strategy=options.update_strategy or project.config["strategy.update"],
            unconstrained=options.unconstrained,
            top=options.top,
            dry_run=options.dry_run,
            packages=options.packages,
            sync=options.sync,
            no_editable=options.no_editable,
            no_self=options.no_self,
            prerelease=options.prerelease,
            fail_fast=options.fail_fast,
            hooks=HookManager(project, options.skip),
        )

    @staticmethod
    def do_update(
        project: Project,
        *,
        selection: GroupSelection,
        strategy: str = "reuse",
        save: str = "compatible",
        unconstrained: bool = False,
        top: bool = False,
        dry_run: bool = False,
        packages: Collection[str] = (),
        sync: bool = True,
        no_editable: bool = False,
        no_self: bool = False,
        prerelease: bool | None = None,
        fail_fast: bool = False,
        hooks: HookManager | None = None,
    ) -> None:
        """Update specified packages or all packages"""
        from itertools import chain

        from pdm.cli.actions import do_lock, do_sync
        from pdm.cli.utils import check_project_file, save_version_specifiers
        from pdm.models.specifiers import get_specifier
        from pdm.utils import normalize_name

        hooks = hooks or HookManager(project)
        check_project_file(project)
        if len(packages) > 0 and (top or len(selection.groups) > 1 or not selection.default):
            raise PdmUsageError("packages argument can't be used together with multiple -G or --no-default or --top.")
        all_dependencies = project.all_dependencies
        updated_deps: dict[str, list[Requirement]] = defaultdict(list)
        locked_groups = project.lockfile.groups
        tracked_names: set[str] = set()
        if not packages:
            if prerelease is not None:
                raise PdmUsageError("--prerelease/--stable must be used with packages given")
            selection.validate()
            for group in selection:
                updated_deps[group] = list(all_dependencies[group])
            tracked_names.update(r.identify() for deps in updated_deps.values() for r in deps)
        else:
            group = selection.one()
            if locked_groups and group not in locked_groups:
                raise ProjectError(f"Requested group not in lockfile: {group}")
            dependencies = all_dependencies[group]
            for name in packages:
                normalized_name = normalize_name(name)
                matched_reqs = [d for d in dependencies if (d.key or "") == normalized_name]
                if not matched_reqs:
                    candidates = project.get_locked_repository().all_candidates
                    if normalized_name not in candidates:
                        raise ProjectError(
                            f"[req]{name}[/] does not exist in [primary]{group}[/] "
                            f"{'dev-' if selection.dev else ''}dependencies nor is a transitive dependency."
                        )
                    look_in_other_group = next(
                        (
                            g
                            for g, deps in all_dependencies.items()
                            if any(normalized_name == d.key for d in deps) and g != group
                        ),
                        None,
                    )
                    if look_in_other_group is not None:
                        raise ProjectError(
                            f"[req]{name}[/] does not exist in [primary]{group}[/], but exists in [primary]{look_in_other_group}[/]."
                            " Please specify the correct group with `-G/--group`."
                        )
                    tracked_names.add(normalized_name)
                else:
                    for req in matched_reqs:
                        req.prerelease = prerelease
                    updated_deps[group].extend(matched_reqs)
            tracked_names.update(r.identify() for deps in updated_deps.values() for r in deps)
            project.core.ui.echo(
                f"Updating {'[bold]global[/] ' if project.is_global else ''}packages: {', '.join(f'[req]{v}[/]' for v in tracked_names)}."
            )
        if unconstrained:
            for deps in updated_deps.values():
                for dep in deps:
                    dep.specifier = get_specifier("")
        reqs = [r for g, deps in all_dependencies.items() for r in deps if locked_groups is None or g in locked_groups]
        # Since dry run is always true in the locking,
        # we need to emit the hook manually with the real dry_run value
        hooks.try_emit("pre_lock", requirements=reqs, dry_run=dry_run)
        with hooks.skipping("pre_lock", "post_lock"):
            resolved = do_lock(
                project,
                strategy=strategy,
                tracked_names=tracked_names,
                requirements=reqs,
                dry_run=True,
                hooks=hooks,
                groups=locked_groups,
            )
        if unconstrained:
            # Need to update version constraints
            save_version_specifiers(chain.from_iterable(updated_deps.values()), resolved, save)
        if not dry_run:
            if unconstrained:
                for group, deps in updated_deps.items():
                    project.add_dependencies(deps, group, selection.dev or False)
            project.write_lockfile(show_message=False)
        hooks.try_emit("post_lock", resolution=resolved, dry_run=dry_run)
        if sync or dry_run:
            do_sync(
                project,
                selection=selection,
                clean=False,
                dry_run=dry_run,
                requirements=[r for deps in updated_deps.values() for r in deps],
                tracked_names=[r.identify() for deps in updated_deps.values() for r in deps] if top else None,
                no_editable=no_editable,
                no_self=no_self or "default" not in selection,
                fail_fast=fail_fast,
                hooks=hooks,
            )



import argparse
import os
from pathlib import Path
from typing import Any, Mapping

from pdm import termui
from pdm._types import RepositoryConfig
from pdm.cli.commands.base import BaseCommand
from pdm.exceptions import PdmUsageError
from pdm.project import Project
from pdm.project.config import DEFAULT_REPOSITORIES, REPOSITORY, SOURCE, Config


class Command(BaseCommand):
    """Display the current configuration"""

    ui: termui.UI

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "-l",
            "--local",
            action="store_true",
            help="Set config in the project's local configuration file",
        )
        parser.add_argument("-d", "--delete", action="store_true", help="Unset a configuration key")
        parser.add_argument(
            "-e",
            "--edit",
            action="store_true",
            help="Edit the configuration file in the default editor(defined by EDITOR env var)",
        )
        parser.add_argument("key", help="Config key", nargs="?")
        parser.add_argument("value", help="Config value", nargs="?")

    @staticmethod
    def get_editor() -> str:
        for key in "VISUAL", "EDITOR":
            rv = os.getenv(key)
            if rv:
                return rv
        if os.name == "nt":
            return "notepad"
        for editor in "sensible-editor", "vim", "nano":
            if os.system(f"which {editor} >/dev/null 2>&1") == 0:
                return editor
        return "vi"

    def edit_file(self, path: Path) -> None:
        import subprocess

        editor = self.get_editor()
        proc = subprocess.Popen(f'{editor} "{path}"', shell=True)

        if proc.wait() != 0:
            raise PdmUsageError(f"Editor {editor} exited abnormally")

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        self.ui = project.core.ui
        if options.edit:
            if options.key:
                raise PdmUsageError("Cannot specify an argument when `--edit` is given")
            if options.delete:
                raise PdmUsageError("`--delete` doesn't work when `--edit` is given")
            config = project.project_config if options.local else project.global_config
            config_path = config.config_file
            config_path.parent.mkdir(parents=True, exist_ok=True)
            return self.edit_file(config_path)
        if options.delete:
            self._delete_config(project, options)
        elif options.value:
            self._set_config(project, options)
        elif options.key:
            self._get_config(project, options)
        else:
            self._list_config(project, options)

    def _get_config(self, project: Project, options: argparse.Namespace) -> None:
        from findpython import ALL_PROVIDERS

        if options.key in project.project_config.deprecated:  # pragma: no cover
            project.core.ui.warn(
                f"[warning]DEPRECATED:[/] the config has been renamed to {project.project_config.deprecated[options.key]}",
            )
            options.key = project.project_config.deprecated[options.key]
        try:
            value = project.project_config[options.key]
        except KeyError:
            value = project.global_config[options.key]
        if options.key.endswith(".password"):
            value = "[i]<hidden>[/i]"
        elif options.key == "python.providers" and not value:
            value = ["venv", *ALL_PROVIDERS]
        project.core.ui.echo(value)

    def _set_config(self, project: Project, options: argparse.Namespace) -> None:
        config = project.project_config if options.local else project.global_config
        if options.key in config.deprecated:  # pragma: no cover
            project.core.ui.warn(
                f"[warning]DEPRECATED:[/] the config has been renamed to {config.deprecated[options.key]}",
            )
        config[options.key] = options.value

    def _show_config(self, config: Mapping[str, Any], supersedes: Mapping[str, Any]) -> None:
        from findpython import ALL_PROVIDERS

        assert Config.site is not None
        for key in sorted(config):
            deprecated = ""
            canonical_key = key
            superseded = key in supersedes
            if key in Config.site.deprecated:  # pragma: no cover
                canonical_key = Config.site.deprecated[key]
                if canonical_key in supersedes:
                    superseded = True
                deprecated = f"[error](deprecating: {key})[/]"
            elif key not in Config._config_map and not (key.startswith("pypi.") or key.startswith(REPOSITORY)):
                continue
            extra_style = "dim" if superseded else None
            if canonical_key not in Config._config_map:
                prefix, name = key.split(".", 1)
                if prefix in (SOURCE, REPOSITORY):
                    title = "non-default PyPI index" if prefix == SOURCE else "custom repository"
                    self.ui.echo(
                        f"[warning]# Configuration of {title} `{name}`",
                        style=extra_style,
                        verbosity=termui.Verbosity.DETAIL,
                    )
                    repository = RepositoryConfig(**config[key], config_prefix=prefix, name=name)
                    if not repository.url and name in DEFAULT_REPOSITORIES:
                        repository.url = DEFAULT_REPOSITORIES[name]
                    self.ui.echo(repository)
                continue
            config_item = Config._config_map[canonical_key]
            self.ui.echo(
                f"[warning]# {config_item.description}",
                style=extra_style,
                verbosity=termui.Verbosity.DETAIL,
            )
            if key.endswith("password"):
                value: Any = "[i]<hidden>[/i]"
            else:
                value = config[key]
                if key == "python.providers" and not value:
                    value = ["venv", *ALL_PROVIDERS]
            self.ui.echo(
                f"[primary]{canonical_key}[/]{deprecated} = {value}",
                style=extra_style,
            )

    def _list_config(self, project: Project, options: argparse.Namespace) -> None:
        assert Config.site is not None
        site_title = "Site/default configuration"
        if Config.site.config_file.exists():
            site_title += f" ([success]{Config.site.config_file}[/])"
        self.ui.echo(site_title, style="bold")
        self._show_config(
            Config.get_defaults(),
            {**project.global_config.self_data, **project.project_config.self_data},
        )

        if project.global_config.self_data:
            self.ui.echo(
                f"\nHome configuration ([success]{project.global_config.config_file}[/]):",
                style="bold",
            )
            self._show_config(project.global_config.self_data, project.project_config.self_data)

        if project.project_config.self_data:
            self.ui.echo(
                f"\nProject configuration ([success]{project.project_config.config_file}[/]):",
                style="bold",
            )
            self._show_config(project.project_config.self_data, {})

    def _delete_config(self, project: Project, options: argparse.Namespace) -> None:
        config = project.project_config if options.local else project.global_config
        if options.key in config.deprecated:  # pragma: no cover
            project.core.ui.warn(
                f"[warning]DEPRECATED:[/] the config has been renamed to {config.deprecated[options.key]}",
            )
        del config[options.key]



from __future__ import annotations

import dataclasses
import itertools
import posixpath
from functools import cached_property
from pathlib import Path
from typing import TYPE_CHECKING, Collection, cast

from dep_logic.markers import AnyMarker, BaseMarker

from pdm.cli.utils import normalize_name
from pdm.exceptions import CandidateNotFound, PdmException
from pdm.models.candidates import Candidate
from pdm.models.markers import EnvSpec, exclude_multi, get_marker
from pdm.models.repositories.base import BaseRepository, CandidateMetadata
from pdm.models.requirements import FileRequirement, Requirement, parse_line
from pdm.utils import cd, url_to_path, url_without_fragments

if TYPE_CHECKING:
    from typing import Any, Callable, Iterable, Mapping

    from pdm._types import FileHash, RepositoryConfig
    from pdm.environments import BaseEnvironment

    CandidateKey = tuple[str, str | None, str | None, bool]


@dataclasses.dataclass(frozen=True)
class Package:
    candidate: Candidate
    dependencies: list[str] | None = None
    summary: str = ""
    marker: BaseMarker = dataclasses.field(default_factory=AnyMarker)


class LockedRepository(BaseRepository):
    def __init__(
        self,
        lockfile: Mapping[str, Any],
        sources: list[RepositoryConfig],
        environment: BaseEnvironment,
        env_spec: EnvSpec | None = None,
    ) -> None:
        super().__init__(sources, environment, env_spec=env_spec or environment.spec)
        self.packages: dict[CandidateKey, Package] = {}
        self.targets: list[EnvSpec] = []
        self._read_lockfile(lockfile)

    def add_package(self, package: Package) -> None:
        self.packages[self._identify_candidate(package.candidate)] = package

    @cached_property
    def all_candidates(self) -> dict[str, list[Candidate]]:
        """Return a dict of all candidates grouped by the package name."""
        result: dict[str, list[Candidate]] = {}
        for entry in self.packages.values():
            result.setdefault(entry.candidate.identify(), []).append(entry.candidate)
        return result

    @property
    def candidates(self) -> dict[str, Candidate]:
        """Return a dict of candidates for the current environment."""
        result: dict[str, Candidate] = {}
        for candidates in self.all_candidates.values():
            for can in candidates:
                if can.req.marker:
                    marker = exclude_multi(can.req.marker, "extras", "dependency_groups")
                    if not marker.matches(self.env_spec):
                        continue
                result[can.identify()] = can
        return result

    def _read_lockfile(self, lockfile: Mapping[str, Any]) -> None:
        if lockfile.get("lock-version"):
            return self._read_pylock(lockfile)
        else:
            return self._read_pdm_lock(lockfile)

    def _read_pylock(self, lockfile: Mapping[str, Any]) -> None:
        root = self.environment.project.root

        if "targets" in lockfile.get("tool", {}).get("pdm", {}):
            self.targets = [EnvSpec.from_spec(**t) for t in lockfile["tool"]["pdm"]["targets"]]
        else:
            for marker in lockfile.get("environments", []):  # pragma: no cover
                self.targets.append(EnvSpec.from_marker(get_marker(cast(str, marker))))

        with cd(root):
            for package in lockfile.get("packages", []):
                group_marker = AnyMarker()
                package_name = package.pop("name")
                req_dict: dict[str, str | bool] = {}
                if "version" in package:
                    req_dict["version"] = f"=={package['version']}"
                if "marker" in package:
                    package_marker = get_marker(cast(str, package["marker"]))
                    req_marker = exclude_multi(package_marker, "extras", "dependency_groups")
                    group_marker = package_marker.inner.only("extras", "dependency_groups")
                    if not req_marker.is_any():
                        req_dict["marker"] = str(req_marker)
                if vcs := package.get("vcs"):  # pragma: no cover
                    req_dict[vcs["type"]] = vcs["url"]
                    req_dict["ref"] = vcs.get("requested-revision")
                    req_dict["revision"] = vcs.get("commit-id")
                    req_dict["subdirectory"] = vcs.get("subdirectory")
                elif directory := package.get("directory"):  # pragma: no cover
                    req_dict["path"] = directory["path"]
                    req_dict["editable"] = directory.get("editable", False)
                    req_dict["subdirectory"] = directory.get("subdirectory")
                elif archive := package.get("archive"):  # pragma: no cover
                    req_dict["url"] = archive.get("url")
                    req_dict["path"] = archive.get("path")
                    req_dict["subdirectory"] = archive.get("subdirectory")
                req = Requirement.from_req_dict(package_name, req_dict)
                if req.is_file_or_url and req.path and not req.url:  # type: ignore[attr-defined]
                    req.url = root.joinpath(req.path).as_uri()  # type: ignore[attr-defined]
                candidate = Candidate(req=req, name=package_name, version=package.get("version"))
                candidate.requires_python = package.get("requires-python", "")
                for artifact in itertools.chain(
                    package.get("wheels", []), [sdist] if (sdist := package.get("sdist")) else []
                ):
                    algo, hash_value = next(iter(artifact["hashes"].items()))
                    hash_item: FileHash = {"hash": f"{algo}:{hash_value}"}
                    if "url" in artifact:
                        hash_item["url"] = artifact["url"]
                    if "name" in artifact:
                        hash_item["file"] = artifact["name"]
                    candidate.hashes.append(hash_item)
                dependencies = package.get("tool", {}).get("pdm", {}).get("dependencies")
                self.packages[self._identify_candidate(candidate)] = Package(candidate, dependencies, "", group_marker)

    def _read_pdm_lock(self, lockfile: Mapping[str, Any]) -> None:
        from pdm.project.lockfile import FLAG_CROSS_PLATFORM, FLAG_STATIC_URLS

        root = self.environment.project.root
        static_urls = FLAG_STATIC_URLS in self.environment.project.lockfile.strategy
        self.targets = [EnvSpec.from_spec(**t) for t in lockfile.get("metadata", {}).get("targets", [])]
        if not self.targets and lockfile:  # pragma: no cover
            # XXX: for reading old lockfiles, to be removed in the future
            if FLAG_CROSS_PLATFORM in self.environment.project.lockfile.strategy:
                self.targets.append(self.environment.allow_all_spec)
            else:
                self.targets.append(self.environment.spec)
        with cd(root):
            for package in lockfile.get("package", []):
                version = package.get("version")
                if version:
                    package["version"] = f"=={version}"
                package_name = package.pop("name")
                req_dict = {
                    k: v
                    for k, v in package.items()
                    if k not in ("dependencies", "requires_python", "summary", "files", "targets")
                }
                req = Requirement.from_req_dict(package_name, req_dict)
                if req.is_file_or_url and req.path and not req.url:  # type: ignore[attr-defined]
                    req.url = root.joinpath(req.path).as_uri()  # type: ignore[attr-defined]
                can = Candidate(req, name=package_name, version=version)
                can.hashes = package.get("files", [])
                if not static_urls and any("url" in f for f in can.hashes):
                    raise PdmException(
                        "Static URLs are not allowed in lockfile unless enabled by `pdm lock --static-urls`."
                    )
                can_id = self._identify_candidate(can)
                can.requires_python = package.get("requires_python", "")
                entry = Package(
                    can,
                    package.get("dependencies", []),
                    package.get("summary", ""),
                )
                self.packages[can_id] = entry

    def _identify_candidate(self, candidate: Candidate) -> CandidateKey:
        url: str | None = None
        if not candidate.req.is_named and candidate.link is not None:
            url = candidate.link.url_without_fragment
            url = self.environment.project.backend.expand_line(cast(str, url))
            if url.startswith("file://"):
                path = posixpath.normpath(url_to_path(url))
                url = Path(path).as_uri()
        return (
            candidate.identify(),
            candidate.version if not url else None,
            url,
            candidate.req.editable,
        )

    def _get_dependencies_from_lockfile(self, candidate: Candidate) -> CandidateMetadata:
        err = (
            f"Missing package {candidate.identify()} from the lockfile, "
            "the lockfile may be broken. Run `pdm lock --update-reuse` to fix it."
        )
        try:
            entry = self.packages[self._identify_candidate(candidate)]
        except KeyError as e:  # pragma: no cover
            raise CandidateNotFound(err) from e

        if entry.dependencies is None:
            raise CandidateNotFound(f"Missing dependencies from the lockfile for package {candidate.identify()}")
        # populate candidate metadata
        if not candidate.name:
            candidate.name = entry.candidate.name
        if not candidate.version:
            candidate.version = entry.candidate.version
        if not candidate.requires_python:
            candidate.requires_python = entry.candidate.requires_python
        deps: list[Requirement] = []
        for line in entry.dependencies:
            deps.append(parse_line(line))
        return CandidateMetadata(deps, candidate.requires_python, entry.summary)

    def dependency_generators(self) -> Iterable[Callable[[Candidate], CandidateMetadata]]:
        return (self._get_dependencies_from_lockfile,)

    def _matching_entries(self, requirement: Requirement) -> Iterable[Package]:
        for key, entry in self.packages.items():
            can_req = entry.candidate.req
            if requirement.name:
                if key[0] != requirement.identify():
                    continue
            else:
                assert isinstance(requirement, FileRequirement)
                if not isinstance(can_req, FileRequirement):
                    continue
                if requirement.path and can_req.path:
                    if requirement.path != can_req.path:
                        continue
                elif key[2] is not None and key[2] != url_without_fragments(requirement.url):
                    continue

            yield entry

    def find_candidates(
        self,
        requirement: Requirement,
        allow_prereleases: bool | None = None,
        ignore_requires_python: bool = False,
        minimal_version: bool = False,
    ) -> Iterable[Candidate]:
        if self.is_this_package(requirement):
            candidate = self.make_this_candidate(requirement)
            if candidate is not None:
                yield candidate
                return
        for entry in self._matching_entries(requirement):
            can = entry.candidate.copy_with(requirement)
            if not requirement.name:
                # make sure can.identify() won't return a randomly-generated name
                requirement.name = can.name
            yield can

    def get_hashes(self, candidate: Candidate) -> list[FileHash]:
        return candidate.hashes

    def evaluate_candidates(self, groups: Collection[str], evaluate_markers: bool = True) -> Iterable[Package]:
        extras, dependency_groups = self.environment.project.split_extras_groups(list(groups))
        excludes = {normalize_name(k) for k in self.environment.project.pyproject.resolution.get("excludes", [])}
        for package in self.packages.values():
            can = package.candidate
            if package.candidate.req.key in excludes:
                continue
            if evaluate_markers and can.req.marker and not can.req.marker.matches(self.env_spec):
                continue
            if not package.marker.evaluate({"extras": set(extras), "dependency_groups": set(dependency_groups)}):
                continue
            if can.req.groups and not any(g in can.req.groups for g in groups):
                continue
            yield package

    def merge_result(self, env_spec: EnvSpec, result: Iterable[Package]) -> None:
        if env_spec not in self.targets:
            self.targets.append(env_spec)
        for entry in result:
            key = self._identify_candidate(entry.candidate)
            existing = self.packages.get(key)
            if existing is None:
                self.packages[key] = entry
            else:
                # merge markers
                old_marker = existing.candidate.req.marker
                if old_marker is None or entry.candidate.req.marker is None:
                    new_marker = None
                else:
                    new_marker = old_marker | entry.candidate.req.marker
                    bare_marker, py_spec = new_marker.split_pyspec()
                    if py_spec.is_superset(self.environment.python_requires):
                        new_marker = bare_marker
                    if new_marker.is_any():
                        new_marker = None
                # merge groups
                new_groups = list(set(existing.candidate.req.groups) | set(entry.candidate.req.groups))
                existing.candidate.req = dataclasses.replace(
                    existing.candidate.req, marker=new_marker, groups=new_groups
                )
                # merge file hashes
                for file in entry.candidate.hashes:
                    if file not in existing.candidate.hashes:
                        existing.candidate.hashes.append(file)
        # clear caches
        if "all_candidates" in self.__dict__:
            del self.__dict__["all_candidates"]



from __future__ import annotations

import abc
import enum
from functools import cached_property
from typing import TYPE_CHECKING, AbstractSet, Any, Iterable, Mapping

import tomlkit

from pdm import termui
from pdm.exceptions import PdmUsageError
from pdm.project.toml_file import TOMLFile

if TYPE_CHECKING:
    from pdm.models.repositories import LockedRepository

GENERATED_COMMENTS = [
    "This file is @generated by PDM.",
    "It is not intended for manual editing.",
]
FLAG_STATIC_URLS = "static_urls"
FLAG_CROSS_PLATFORM = "cross_platform"
FLAG_DIRECT_MINIMAL_VERSIONS = "direct_minimal_versions"
FLAG_INHERIT_METADATA = "inherit_metadata"


class Compatibility(enum.IntEnum):
    NONE = 0  # The lockfile can't be read by the current version of PDM.
    SAME = 1  # The lockfile version is the same as the current version of PDM.
    BACKWARD = 2  # The current version of PDM is newer than the lockfile version.
    FORWARD = 3  # The current version of PDM is older than the lockfile version.


class Lockfile(TOMLFile, metaclass=abc.ABCMeta):
    SUPPORTED_FLAGS: AbstractSet[str]

    @property
    @abc.abstractmethod
    def hash(self) -> tuple[str, str]:
        """The content hash algo and hash value of the pyproject.toml to generate this lockfile."""
        pass

    @abc.abstractmethod
    def update_hash(self, hash_value: str, algo: str = "sha256") -> None:
        """Update the content hash of the lockfile."""
        pass

    @property
    @abc.abstractmethod
    def groups(self) -> list[str] | None:
        """The groups defined in the lockfile, or None if no groups are defined."""
        return []

    @property
    @abc.abstractmethod
    def strategy(self) -> set[str]:
        """The strategies used in the lockfile."""
        return set()

    @cached_property
    @abc.abstractmethod
    def default_strategies(self) -> set[str]:
        """The default strategies to be used if no strategies are defined in the lockfile."""
        return set()

    def apply_strategy_change(self, changes: Iterable[str]) -> set[str]:
        """Apply the given strategy changes to the current strategy."""
        original = self.strategy
        supported = self.SUPPORTED_FLAGS
        for change in changes:
            change = change.replace("-", "_").lower()
            if change.startswith("no_"):
                if change[3:] not in supported:
                    raise PdmUsageError(f"Invalid strategy flag: {change[3:]}, supported: {', '.join(supported)}")
                original.discard(change[3:])
            else:
                if change not in supported:
                    raise PdmUsageError(f"Invalid strategy flag: {change}, supported: {', '.join(supported)}")
                original.add(change)
        return original

    def compare_groups(self, groups: Iterable[str]) -> list[str]:
        """Compare the given groups with the lockfile groups and return the groups that are not in the lockfile."""
        if not self.groups:
            return []
        return list(set(groups).difference(self.groups))

    def set_data(self, data: Mapping[str, Any]) -> None:
        doc = tomlkit.document()
        for line in GENERATED_COMMENTS:
            doc.append(None, tomlkit.comment(line))
        doc.update(data)
        super().set_data(doc)

    def write(self, show_message: bool = True) -> None:
        super().write()
        if show_message:
            self.ui.echo(f"Changes are written to [success]{self._path.name}[/].", verbosity=termui.Verbosity.NORMAL)

    def __getitem__(self, key: str) -> dict:
        return self._data[key]  # type: ignore[return-value]

    @abc.abstractmethod
    def compatibility(self) -> Compatibility:
        """We use a three-part versioning scheme for lockfiles:
        The first digit represents backward compatibility and the second digit represents forward compatibility.
        """

    @abc.abstractmethod
    def format_lockfile(self, repository: LockedRepository, groups: Iterable[str] | None, strategy: set[str]) -> None:
        """Format lock file from a dict of resolved candidates, a mapping of dependencies
        and a collection of package summaries.
        """



from __future__ import annotations

import os
import sqlite3
import sys
import threading
from contextlib import closing
from functools import lru_cache
from pathlib import Path
from typing import TYPE_CHECKING, Any, cast

import hishel
import hishel.httpx
import httpx
from unearth.fetchers import PyPIClient

from pdm.__version__ import __version__
from pdm.termui import logger

if TYPE_CHECKING:
    from ssl import SSLContext

    from pdm._types import RepositoryConfig


def _create_truststore_ssl_context() -> SSLContext | None:
    if sys.version_info < (3, 10):
        return None

    try:
        import ssl
    except ImportError:
        return None

    try:
        import truststore
    except ImportError:
        return None

    import certifi

    ctx = truststore.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
    ctx.load_verify_locations(certifi.where())
    return ctx


_ssl_context = _create_truststore_ssl_context()
CACHES_TTL = 7 * 24 * 60 * 60  # 7 days
MAX_RETRIES = 4


@lru_cache(maxsize=None)
def _get_transport(
    verify: bool | SSLContext | str = True,
    cert: tuple[str, str | None] | None = None,
    proxy: httpx.Proxy | None = None,
) -> httpx.BaseTransport:
    return httpx.HTTPTransport(verify=verify, cert=cert, trust_env=True, proxy=proxy, retries=MAX_RETRIES)


class ThreadedSyncSqliteStorage(hishel.SyncSqliteStorage):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        self._local_conns: dict[int, sqlite3.Connection] = {}
        self._lock = threading.Lock()
        self._initialized = False
        super().__init__(*args, **kwargs)

    @property
    def connection(self) -> sqlite3.Connection | None:
        return self._local_conns.get(threading.get_ident())

    @connection.setter
    def connection(self, conn: sqlite3.Connection | None) -> None:
        if conn is not None:
            self._local_conns[threading.get_ident()] = conn

    def close(self) -> None:
        with self._lock:
            while self._local_conns:
                _, conn = self._local_conns.popitem()
                conn.close()

    def _ensure_connection(self) -> sqlite3.Connection:
        """
        Ensure connection is established and database is initialized.

        We need to create connections with check_same_thread=False for the sake of close(),
        so we have to open-code the entire implementation here.
        """

        # we take a lock _despite_ using TLS to protect against concurrent close(), otherwise we have a TOCTOU
        # this is kinda ugly and defeats the purpose of using TLS, but eh
        with self._lock:
            if self.connection is None:
                # Create cache directory and resolve full path on first connection
                self.database_path.parent.mkdir(parents=True, exist_ok=True)
                full_path = self.database_path.resolve()
                conn = sqlite3.connect(str(full_path), check_same_thread=False)
                with closing(conn.cursor()) as cursor:
                    cursor.execute("PRAGMA foreign_keys=ON")
                self.connection = conn
            if not self._initialized:
                self._initialize_database()
                self._initialized = True
            return self.connection


class PDMPyPIClient(PyPIClient):
    def __init__(self, *, sources: list[RepositoryConfig], cache_dir: Path | None = None, **kwargs: Any) -> None:
        import shutil

        from httpx._utils import URLPattern
        from unearth.fetchers.sync import LocalFSTransport

        if cache_dir is None:

            def cache_transport(transport: httpx.BaseTransport) -> httpx.BaseTransport:
                return transport
        else:
            # clean up old (pre-hishel 1.0) cache
            cache_db = cache_dir / "http-cache.db"
            if not cache_db.exists():
                for f in cache_dir.iterdir():
                    if not f.name.startswith("http-cache.db"):
                        if f.is_dir():
                            shutil.rmtree(f, ignore_errors=True)
                        else:
                            f.unlink()
            storage = ThreadedSyncSqliteStorage(database_path=cache_db, default_ttl=CACHES_TTL)

            def cache_transport(transport: httpx.BaseTransport) -> httpx.BaseTransport:
                return hishel.httpx.SyncCacheTransport(next_transport=transport, storage=storage)

        mounts: dict[str, httpx.BaseTransport] = {"file://": LocalFSTransport()}
        self._trusted_host_ports: set[tuple[str, int | None]] = set()
        self._proxy_map = {
            URLPattern(key): proxy for key, proxy in self._get_proxy_map(None, allow_env_proxies=True).items()
        }
        self._proxy_map = dict(sorted(self._proxy_map.items()))
        for s in sources:
            assert s.url is not None
            url = httpx.URL(s.url)
            if s.verify_ssl is False:
                self._trusted_host_ports.add((url.host, url.port))
            if s.name == "pypi":
                kwargs["transport"] = self._transport_for(s)
                continue
            mounts[f"{url.scheme}://{url.netloc.decode('ascii')}/"] = cache_transport(self._transport_for(s))
        mounts.update(kwargs.pop("mounts", None) or {})
        kwargs.update(follow_redirects=True)

        httpx.Client.__init__(self, mounts=mounts, **kwargs)

        self.headers["User-Agent"] = self._make_user_agent()
        self.event_hooks["response"].append(self.on_response)
        self._transport = cache_transport(self._transport)  # type: ignore[has-type]

    def _transport_for(self, source: RepositoryConfig) -> httpx.BaseTransport:
        if source.verify_ssl is False:
            verify: str | bool | SSLContext = False
        elif source.ca_certs:
            verify = source.ca_certs
        else:
            verify = os.getenv("REQUESTS_CA_BUNDLE") or os.getenv("CURL_CA_BUNDLE") or _ssl_context or True
        if source.client_cert:
            cert = (source.client_cert, source.client_key)
        else:
            cert = None
        source_url = httpx.URL(cast(str, source.url))
        proxy = next((proxy for pattern, proxy in self._proxy_map.items() if pattern.matches(source_url)), None)
        return _get_transport(verify=verify, cert=cert, proxy=proxy)

    def _make_user_agent(self) -> str:
        import platform

        return f"pdm/{__version__} {platform.python_implementation()}/{platform.python_version()} {platform.system()}/{platform.release()}"

    def on_response(self, response: httpx.Response) -> None:
        from unearth.utils import ARCHIVE_EXTENSIONS

        if response.extensions.get("from_cache"):
            response.from_cache = True  # type: ignore[attr-defined]
            if response.url.path.endswith(ARCHIVE_EXTENSIONS):
                logger.info("Using cached response for %s", response.url)



from __future__ import annotations

import argparse
from typing import TYPE_CHECKING, Any, cast

from pdm import termui
from pdm.cli import actions
from pdm.cli.commands.base import BaseCommand
from pdm.cli.hooks import HookManager
from pdm.cli.options import skip_option
from pdm.cli.templates import ProjectTemplate
from pdm.exceptions import PdmUsageError, ProjectError
from pdm.models.backends import _BACKENDS, DEFAULT_BACKEND, BuildBackend, get_backend
from pdm.models.specifiers import get_specifier
from pdm.utils import (
    get_user_email_from_git,
    package_installed,
    sanitize_project_name,
    validate_project_name,
)

if TYPE_CHECKING:
    from pdm.project import Project


class Command(BaseCommand):
    """Initialize a pyproject.toml for PDM.

    Built-in templates:
    - default: `pdm init`, A simple template with a basic structure.
    - minimal: `pdm init minimal`, A minimal template with only `pyproject.toml`.
    """

    supports_other_generator = True

    def __init__(self) -> None:
        self.interactive = True

    def initialize_git(self, project: Project) -> None:
        """Initialize a git repository if git is available and .git doesn't exist."""
        import shutil
        import subprocess

        if (project.root / ".git").exists():
            project.core.ui.info("Git repository already exists, skipping initialization.")
            return

        git_command = shutil.which("git")
        if not git_command:
            project.core.ui.info("Git command not found, skipping initialization.")
            return

        try:
            subprocess.run(
                [git_command, "init"],
                cwd=project.root,
                check=True,
                capture_output=True,
                encoding="utf-8",
            )
            project.core.ui.info("Git repository initialized successfully.")
        except subprocess.CalledProcessError as e:
            project.core.ui.error(f"Failed to initialize Git repository: {e.stderr}")

    def do_init(self, project: Project, options: argparse.Namespace) -> None:
        """Bootstrap the project and create a pyproject.toml"""
        hooks = HookManager(project, options.skip)
        if options.generator == "copier":
            self._init_copier(project, options)
        elif options.generator == "cookiecutter":
            self._init_cookiecutter(project, options)
        else:
            self.set_python(project, options.python, hooks)
            self._init_builtin(project, options)

        if options.init_git:
            self.initialize_git(project)

        hooks.try_emit("post_init")

    def _init_copier(self, project: Project, options: argparse.Namespace) -> None:
        if not package_installed("copier"):
            raise PdmUsageError(
                "--copier is passed but copier is not installed. Install it by `pdm self add copier`"
            ) from None

        from copier.cli import CopierApp

        if not options.template:
            raise PdmUsageError("template argument is required when --copier is passed")
        _, retval = CopierApp.run(
            ["copier", "copy", options.template, str(project.root), *options.generator_args], exit=False
        )
        if retval != 0:
            raise RuntimeError("Copier exited with non-zero status code")

    def _init_cookiecutter(self, project: Project, options: argparse.Namespace) -> None:
        if not package_installed("cookiecutter"):
            raise PdmUsageError(
                "--cookiecutter is passed but cookiecutter is not installed. Install it by `pdm self add cookiecutter`"
            ) from None

        from cookiecutter.cli import main as cookiecutter

        if not options.template:
            raise PdmUsageError("template argument is required when --cookiecutter is passed")
        try:
            cookiecutter.main(
                [options.template, "--output-dir", str(project.root), *options.generator_args], standalone_mode=False
            )
        except SystemExit as e:
            raise RuntimeError("Cookiecutter exited with an error") from e

    def _init_builtin(self, project: Project, options: argparse.Namespace) -> None:
        metadata = self.get_metadata_from_input(project, options)
        template = options.template
        if not template:
            template = "default" if options.dist else "minimal"
        with ProjectTemplate(template) as template:
            template.generate(project.root, metadata, options.overwrite)
        project.pyproject.reload()

    def set_interactive(self, value: bool) -> None:
        self.interactive = value

    def ask(self, question: str, default: str) -> str:
        if not self.interactive:
            return default
        return termui.ask(question, default=default)

    def ask_project(self, project: Project) -> str:
        default = sanitize_project_name(project.root.name)
        name = self.ask("Project name", default)
        if default == name or validate_project_name(name):
            return name
        project.core.ui.echo(
            "Project name is not valid, it should follow PEP 426",
            err=True,
            style="warning",
        )
        return self.ask_project(project)

    def get_metadata_from_input(self, project: Project, options: argparse.Namespace) -> dict[str, Any]:
        from pdm.formats.base import array_of_inline_tables, make_array, make_inline_table

        if options.name:
            if not validate_project_name(options.name):
                raise ProjectError("Project name is not valid, it should follow PEP 426")
            name = options.name
        else:
            name = self.ask_project(project)
        version = self.ask("Project version", options.project_version or "0.1.0")
        is_dist = options.dist or bool(options.backend)
        if not is_dist and self.interactive:
            is_dist = termui.confirm(
                "Do you want to build this project for distribution(such as wheel)?\n"
                "If yes, it will be installed by default when running `pdm install`."
            )
        options.dist = is_dist
        build_backend: type[BuildBackend] | None = None
        python = project.python
        if is_dist:
            description = self.ask("Project description", "")
            if options.backend:
                build_backend = get_backend(options.backend)
            elif self.interactive:
                all_backends = list(_BACKENDS)
                project.core.ui.echo("Which build backend to use?")
                for i, backend in enumerate(all_backends):
                    project.core.ui.echo(f"{i}. [success]{backend}[/]")
                selected_backend = termui.ask(
                    "Please select",
                    prompt_type=int,
                    choices=[str(i) for i in range(len(all_backends))],
                    show_choices=False,
                    default=0,
                )
                build_backend = get_backend(all_backends[int(selected_backend)])
            else:
                build_backend = DEFAULT_BACKEND
            default_python_requires = f">={python.major}.{python.minor}"
        else:
            description = ""
            default_python_requires = f"=={python.major}.{python.minor}.*"
        license = self.ask("License(SPDX name)", options.license or "MIT")

        git_user, git_email = get_user_email_from_git()
        author = self.ask("Author name", git_user)
        email = self.ask("Author email", git_email)
        python_requires = self.ask("Python requires('*' to allow any)", default_python_requires)

        data = {
            "project": {
                "name": name,
                "version": version,
                "authors": array_of_inline_tables([{"name": author, "email": email}]),
                "license": make_inline_table({"text": license}),
                "dependencies": make_array([], True),
            },
            "tool": {"pdm": {"distribution": is_dist}},
        }

        if python_requires and python_requires != "*":
            get_specifier(python_requires)
            data["project"]["requires-python"] = python_requires  # type: ignore[index]
        if description:
            data["project"]["description"] = description  # type: ignore[index]
        if build_backend is not None:
            data["build-system"] = cast(dict, build_backend.build_system())

        return data

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        skip_option.add_to_parser(parser)

        status = {
            False: termui.style("\\[not installed]", style="error"),
            True: termui.style("\\[installed]", style="success"),
        }
        if self.supports_other_generator:
            generator = parser.add_mutually_exclusive_group()
            generator.add_argument(
                "--copier",
                action="store_const",
                dest="generator",
                const="copier",
                help=f"Use Copier to generate project {status[package_installed('copier')]}",
            )
            generator.add_argument(
                "--cookiecutter",
                action="store_const",
                dest="generator",
                const="cookiecutter",
                help=f"Use Cookiecutter to generate project {status[package_installed('cookiecutter')]}",
            )
        group = parser.add_argument_group("builtin generator options")
        group.add_argument(
            "-n",
            "--non-interactive",
            action="store_true",
            help="Don't ask questions but use default values",
        )
        group.add_argument("--python", help="Specify the Python version/path to use")
        group.add_argument(
            "--dist", "--lib", dest="dist", action="store_true", help="Create a package for distribution"
        )
        group.add_argument("--backend", choices=list(_BACKENDS), help="Specify the build backend, which implies --dist")
        group.add_argument("--license", help="Specify the license (SPDX name)")
        group.add_argument("--name", help="Specify the project name")
        group.add_argument("--project-version", help="Specify the project's version")
        group.add_argument(
            "--no-git", dest="init_git", action="store_false", default=True, help="Do not initialize a git repository"
        )
        parser.add_argument(
            "template", nargs="?", help="Specify the project template, which can be a local path or a Git URL"
        )
        if self.supports_other_generator:
            parser.add_argument("generator_args", nargs=argparse.REMAINDER, help="Arguments passed to the generator")
        parser.add_argument("-r", "--overwrite", action="store_true", help="Overwrite existing files")
        parser.set_defaults(search_parent=False, generator="builtin")

    def set_python(self, project: Project, python: str | None, hooks: HookManager) -> None:
        from pdm.cli.commands.use import Command as UseCommand

        python_info = UseCommand().do_use(
            project,
            python or "",
            first=bool(python) or not self.interactive,
            ignore_remembered=True,
            ignore_requires_python=True,
            save=False,
            hooks=hooks,
        )

        if python_info.get_venv() is None:
            project.core.ui.info(
                "You are using the PEP 582 mode, no virtualenv is created.\n"
                "You can change configuration with `pdm config python.use_venv True`.\n"
                "For more info, please visit https://peps.python.org/pep-0582/"
            )
        project.python = python_info

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        if project.pyproject.exists():
            project.core.ui.echo("pyproject.toml already exists, update it now.", style="primary")
        else:
            project.core.ui.echo("Creating a pyproject.toml for PDM...", style="primary")
        self.set_interactive(not options.non_interactive and termui.is_interactive())
        self.do_init(project, options=options)
        project.core.ui.echo("Project is initialized successfully", style="primary")
        if self.interactive:
            actions.ask_for_import(project)



# Copyright 2024 Ulrik Sverdrup "bluss"
#
# Permission is hereby granted, free of charge, to any person obtaining a copy of
# this software and associated documentation files (the "Software"), to deal in
# the Software without restriction, including without limitation the rights to
# use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
# the Software, and to permit persons to whom the Software is furnished to do so,
# subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
# FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
# COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
# IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
# CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

"""
Patch sysconfigdata and pkgconfig files in
a python installation from indygreg's python builds.

Adapted from https://github.com/bluss/sysconfigpatcher
"""

import ast
import logging
import os
import re
import shutil
from dataclasses import dataclass
from pathlib import Path

_logger = logging.getLogger(__name__)

OLD_PREFIX = "/install"

SYSCONFIG_HEADER = """\
# system configuration generated and used by the sysconfig module
# install path patched by sysconfigpatcher
"""


@dataclass
class WordReplace:
    """Replace word with another"""

    word: str
    to: str


DEFAULT_VARIABLE_UPDATES = {
    "CC": WordReplace("clang", "cc"),
    "CXX": WordReplace("clang++", "c++"),
    "BLDSHARED": WordReplace("clang", "cc"),
    "LDSHARED": WordReplace("clang", "cc"),
    "LDCXXSHARED": WordReplace("clang++", "c++"),
    "LINKCC": WordReplace("clang", "cc"),
    "AR": "ar",
}


def read_sysconfig_data_ast(fname):
    with open(fname) as fn:
        module = ast.parse(fn.read(), filename=fname)
    return module


def update_prefix(value: str, real_prefix: str):
    if value.startswith(OLD_PREFIX):
        return value.replace(OLD_PREFIX, real_prefix, 1)
    return value


def sync_file(fn):
    if hasattr(os, "fdatasync"):
        os.fdatasync(fn)
    elif hasattr(os, "fsync"):
        os.fsync(fn)


def select_child(ast_obj, type_):
    return next((elt for elt in ast_obj.body if isinstance(elt, type_)), None)


def patch_sysconfig_ast(obj, real_prefix, variable_updates=None):
    """
    variable_updates (dict[str, str | WordReplace] | None): Extra variables that should be updated
    return True if any changes were done
    """
    did_update = False
    variable_updates = variable_updates or {}

    # find module body
    if "body" not in obj._fields:
        raise ValueError
    # find assignment
    assignment = select_child(obj, ast.Assign)
    if assignment is None:
        raise ValueError

    dict_ast = assignment.value

    if not isinstance(dict_ast, ast.Dict):
        raise ValueError(f"Expected Dict, got {dict_ast!r}")

    real_prefix_str = str(real_prefix)
    # type check
    if not all(isinstance(key, ast.Constant) and isinstance(key.value, str) for key in dict_ast.keys):
        raise ValueError("Expected all str keys dict")
    if not all(isinstance(value, ast.Constant) and isinstance(value.value, (str, int)) for value in dict_ast.values):
        raise ValueError("Expected all str and int values dict")
    # index because we are modifying
    for key_ast, value_ast in zip(dict_ast.keys, dict_ast.values):
        if not (isinstance(key_ast, ast.Constant) and isinstance(key_ast.value, str)):
            raise ValueError("Expected all str keys dict")
        if not (isinstance(value_ast, ast.Constant) and isinstance(value_ast.value, (str, int))):
            raise ValueError("Expected all str and int values dict")
        key = key_ast.value
        value = value_ast.value

        if not isinstance(value, str):
            continue

        new_value = None

        if key in variable_updates:
            updater = variable_updates[key]
            if isinstance(updater, WordReplace):
                new_value = re.sub(r"(^|\s)" + re.escape(updater.word) + r"(?=\s|$)", updater.to, value, flags=re.ASCII)
            else:
                new_value = updater
            if value == new_value:
                new_value = None
        elif value.startswith(OLD_PREFIX):
            # some keys have multiple paths like
            # DESTDIRS="/install /install/lib /install/lib/python3.12 (...)"
            new_value = " ".join(update_prefix(part, real_prefix_str) for part in value.split(" "))

        if new_value is not None:
            did_update = True
            value_ast.value = new_value
            _logger.debug("Updated %r value\n  from %r\n  to %r", key, value, new_value)

    return did_update


def patch_sysconfig(path: Path, real_prefix: Path, dry_run: bool, backup_files: bool, variable_updates=None):
    """
    return True if did patch or nothing to patch
    """
    _logger.debug("Reading %r", str(path))
    obj = read_sysconfig_data_ast(path)
    did_update = patch_sysconfig_ast(obj, real_prefix, variable_updates)

    if not did_update:
        _logger.info("Nothing to patch in sysconfig")
        return True

    new_file = path.with_suffix(".py.new")
    if dry_run:
        _logger.info("Would patch %s", path)
    else:
        if backup_files:
            backup_file = path.with_suffix(".py.backup")
            shutil.copy(path, backup_file)
            _logger.debug("Wrote %s", backup_file)

        with open(new_file, "w") as fn:
            fn.write(SYSCONFIG_HEADER)
            fn.write(ast.unparse(obj))
            sync_file(fn)

        shutil.move(new_file, path)
        _logger.info("Patched %s", path)
    return True


def find_pkgconfigs(path: Path):
    pkgconfig = path / "lib/pkgconfig"
    if not pkgconfig.exists() or not pkgconfig.is_dir():
        return

    # is_file and not is_symlink
    for child in pkgconfig.iterdir():
        if child.is_file() and not child.is_symlink() and child.suffix == ".pc":
            yield child


def write_new_pkgconfig(fname: Path, real_prefix: Path, dest_path: Path):
    """
    raises ValueError if there is a problem
    returns True if file was changed
    """

    def replace_func(matchobj):
        return matchobj.group(1) + str(real_prefix)

    did_update = False

    with open(fname) as fn:
        with open(dest_path, "w") as outfile:
            for line in fn:
                new_line = re.sub(
                    r"^(\w+=)(" + re.escape(OLD_PREFIX) + ")",
                    replace_func,
                    line,
                    count=1,
                )
                if new_line != line:
                    did_update = True
                    _logger.debug("Updated\n  from %r\n  to %r", line.rstrip(), new_line.rstrip())
                outfile.write(new_line)
            sync_file(outfile)
    return did_update


def patch_pkgconfig(pkgconfig_file, real_prefix, dry_run: bool, backup_files: bool):
    new_file = pkgconfig_file.with_suffix(".pc.new")
    if dry_run:
        _logger.info("Would patch %s", pkgconfig_file)
    else:
        did_update = write_new_pkgconfig(pkgconfig_file, real_prefix, new_file)
        if not did_update:
            os.unlink(new_file)
            _logger.info("Nothing to patch for %s", pkgconfig_file)
        else:
            if backup_files:
                backup_file = pkgconfig_file.with_suffix(".pc.backup")
                shutil.copy(pkgconfig_file, backup_file)
                _logger.debug("Wrote %s", backup_file)
            shutil.move(new_file, pkgconfig_file)
            _logger.info("Patched %s", pkgconfig_file)


def find_libdir(real_prefix: Path):
    # probe python for its libdir and path to the file
    libdir = real_prefix / "lib"
    # find python3.xy in libdir

    if libdir.is_dir():
        py_children = [child for child in libdir.iterdir() if child.is_dir() and child.name.startswith("python3")]
        if len(py_children) == 1:
            return py_children[0]
        if len(py_children) > 1:
            _logger.info("Found lib directories: %r", py_children)
    _logger.error("No lib/python3.x directory found")
    return None


def find_sysconfigdata(real_prefix: Path):
    # probe python for its libdir and path to the file
    sysconfigdata_prefix = "_sysconfigdata_"
    libdir = find_libdir(real_prefix)
    if libdir is None:
        return None
    for child in libdir.iterdir():
        if (
            child.is_file()
            and not child.is_symlink()
            and child.name.startswith(sysconfigdata_prefix)
            and child.suffix == ".py"
        ):
            return child
    return None


def patch(real_prefix: Path) -> None:
    if os.name == "nt":
        # sysconfig patching not needed on Windows
        return
    sysconfig_path = find_sysconfigdata(real_prefix)
    if sysconfig_path is None:
        _logger.error("No sysconfigdata file found")
    else:
        patch_sysconfig(
            sysconfig_path, real_prefix, dry_run=False, backup_files=False, variable_updates=DEFAULT_VARIABLE_UPDATES
        )

    pkgconfig_files = list(find_pkgconfigs(real_prefix))
    if not pkgconfig_files:
        _logger.info("No pkgconfig files found")
    for pc_file in pkgconfig_files:
        patch_pkgconfig(pc_file, real_prefix, dry_run=False, backup_files=False)



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from pdm.formats import flit, pipfile, poetry, requirements, setup_py
from pdm.formats.base import MetaConvertError as MetaConvertError

if TYPE_CHECKING:
    from argparse import Namespace
    from pathlib import Path
    from typing import Iterable, Mapping, Protocol, Union

    from pdm.models.candidates import Candidate
    from pdm.models.requirements import Requirement
    from pdm.project import Project

    ExportItems = Union[Iterable[Candidate], Iterable[Requirement]]

    class _Format(Protocol):
        def check_fingerprint(self, project: Project | None, filename: str | Path) -> bool: ...

        def convert(
            self,
            project: Project | None,
            filename: str | Path,
            options: Namespace | None,
        ) -> tuple[Mapping, Mapping]: ...

        def export(self, project: Project, candidates: ExportItems, options: Namespace | None) -> str: ...


FORMATS: Mapping[str, _Format] = {
    "pipfile": cast("_Format", pipfile),
    "poetry": cast("_Format", poetry),
    "flit": cast("_Format", flit),
    "setuppy": cast("_Format", setup_py),
    "requirements": cast("_Format", requirements),
}



from __future__ import annotations

import argparse
from pathlib import Path
from typing import Iterable

import tomlkit

from pdm.cli.commands.base import BaseCommand
from pdm.cli.filters import GroupSelection
from pdm.cli.options import groups_group, lockfile_option
from pdm.exceptions import PdmUsageError
from pdm.formats import FORMATS
from pdm.formats.pylock import PyLockConverter
from pdm.models.candidates import Candidate
from pdm.models.requirements import Requirement
from pdm.project import Project
from pdm.project.lockfile import FLAG_INHERIT_METADATA


class Command(BaseCommand):
    """Export the locked packages set to other formats"""

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        lockfile_option.add_to_parser(parser)
        parser.add_argument(
            "-f",
            "--format",
            choices=["requirements", "pylock"],
            default="requirements",
            help="Export to requirements.txt format or pylock.toml format",
        )
        groups_group.add_to_parser(parser)
        parser.add_argument(
            "--no-hashes",
            "--without-hashes",
            dest="hashes",
            action="store_false",
            default=True,
            help="Don't include artifact hashes",
        )
        parser.add_argument(
            "--no-markers",
            action="store_false",
            default=True,
            dest="markers",
            help="(DEPRECATED)Don't include platform markers",
        )
        parser.add_argument(
            "--no-extras", action="store_false", default=True, dest="extras", help="Strip extras from the requirements"
        )
        parser.add_argument(
            "-o",
            "--output",
            help="Write output to the given file, or print to stdout if not given",
        )
        parser.add_argument(
            "--pyproject",
            action="store_true",
            help="Read the list of packages from pyproject.toml",
        )
        parser.add_argument("--expandvars", action="store_true", help="Expand environment variables in requirements")
        group = parser.add_mutually_exclusive_group()
        group.add_argument("--self", action="store_true", help="Include the project itself")
        group.add_argument(
            "--editable-self", action="store_true", help="Include the project itself as an editable dependency"
        )

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        from pdm.models.repositories.lock import Package

        if options.format == "pylock":
            locked_repository = project.get_locked_repository()
            if options.self or options.editable_self:
                locked_repository.add_package(
                    Package(project.make_self_candidate(editable=options.editable_self), [], "")
                )
            doc = tomlkit.dumps(PyLockConverter(project, locked_repository).convert())
            if options.output:
                Path(options.output).write_text(doc, encoding="utf-8")
            else:
                print(doc)
            return
        if options.pyproject:
            options.hashes = False
        selection = GroupSelection.from_options(project, options)
        if options.markers is False:
            project.core.ui.warn(
                "The --no-markers option is on, the exported requirements can only work on the current platform"
            )
        packages: Iterable[Requirement] | Iterable[Candidate]
        if options.pyproject:
            all_deps = project._resolve_dependencies(list(selection))
            packages = [r for group in selection for r in project.get_dependencies(group, all_deps)]
        else:
            if not project.lockfile.exists():
                raise PdmUsageError("No lockfile found, please run `pdm lock` first.")
            if FLAG_INHERIT_METADATA not in project.lockfile.strategy:
                raise PdmUsageError(
                    "Can't export a lock file without environment markers, please re-generate the lock file with `inherit_metadata` strategy."
                )
            groups = set(selection)
            candidates = sorted(
                (
                    entry.candidate
                    for entry in project.get_locked_repository().evaluate_candidates(groups, not options.markers)
                ),
                key=lambda c: not c.req.extras,
            )
            packages = []
            seen_extras: set[str] = set()
            for candidate in candidates:
                if options.extras:
                    key = candidate.req.key or ""
                    if candidate.req.extras:
                        seen_extras.add(key)
                    elif key in seen_extras:
                        continue
                elif candidate.req.extras:
                    continue
                if not options.markers and candidate.req.marker:
                    candidate.req.marker = None
                packages.append(candidate)  # type: ignore[arg-type]

        content = FORMATS[options.format].export(project, packages, options)
        if options.output:
            Path(options.output).write_text(content, encoding="utf-8")
        else:
            # Use a regular print to avoid any formatting / wrapping.
            print(content)



from pdm.installers.base import BaseSynchronizer
from pdm.installers.manager import InstallManager
from pdm.installers.synchronizers import Synchronizer
from pdm.installers.uv import UvSynchronizer

__all__ = ["BaseSynchronizer", "InstallManager", "Synchronizer", "UvSynchronizer"]



"""
Utility functions
"""

from __future__ import annotations

import atexit
import contextlib
import functools
import inspect
import json
import os
import re
import shutil
import subprocess
import sys
import sysconfig
import tempfile
import urllib.parse as parse
import warnings
from datetime import datetime, timedelta, timezone
from os import name as os_name
from pathlib import Path
from typing import TYPE_CHECKING, Mapping

from packaging.specifiers import InvalidSpecifier, SpecifierSet
from packaging.version import Version

from pdm.compat import importlib_metadata
from pdm.exceptions import PDMDeprecationWarning, PdmException

if TYPE_CHECKING:
    from re import Match
    from typing import IO, Any, Iterator

    from pbs_installer import PythonVersion

    from pdm._types import FileHash, HiddenText, RepositoryConfig
    from pdm.compat import Distribution

try:
    _packaging_version = importlib_metadata.version("packaging")
except Exception:
    from packaging import __version__ as _packaging_version


@functools.lru_cache(maxsize=1024)
def parse_version(version: str) -> Version:
    return Version(version)


PACKAGING_22 = parse_version(_packaging_version) >= parse_version("22")


def create_tracked_tempdir(suffix: str | None = None, prefix: str | None = None, dir: str | None = None) -> str:
    name = tempfile.mkdtemp(suffix=suffix, prefix=prefix, dir=dir)
    os.makedirs(name, mode=0o777, exist_ok=True)

    def clean_up() -> None:
        shutil.rmtree(name, ignore_errors=True)

    atexit.register(clean_up)
    return name


def get_trusted_hosts(sources: list[RepositoryConfig]) -> list[str]:
    """Parse the project sources and return the trusted hosts"""
    trusted_hosts = []
    for source in sources:
        assert source.url
        url = source.url
        netloc = parse.urlparse(url).netloc
        host = netloc.rsplit("@", 1)[-1]
        if host not in trusted_hosts and source.verify_ssl is False:
            trusted_hosts.append(host)
    return trusted_hosts


def url_without_fragments(url: str) -> str:
    return parse.urlunparse(parse.urlparse(url)._replace(fragment=""))


def join_list_with(items: list[Any], sep: Any) -> list[Any]:
    new_items = []
    for item in items:
        new_items.extend([item, sep])
    return new_items[:-1]


def find_project_root(cwd: str = ".") -> str | None:
    """Recursively find a `pyproject.toml` at given path or current working directory."""
    path = Path(cwd).absolute()
    if list(path.glob("pyproject.toml")):
        return path.as_posix()

    if path == path.parent:
        return None

    return find_project_root(str(path.parent))


def convert_hashes(files: list[FileHash]) -> dict[str, list[str]]:
    """Convert Pipfile.lock hash lines into InstallRequirement option format.

    The option format uses a str-list mapping. Keys are hash algorithms, and
    the list contains all values of that algorithm.
    """
    result: dict[str, list[str]] = {}
    for f in files:
        hash_value = f.get("hash", "")
        name, has_name, hash_value = hash_value.partition(":")
        if not has_name:
            name, hash_value = "sha256", name
        result.setdefault(name, []).append(hash_value)
    return result


def get_user_email_from_git() -> tuple[str, str]:
    """Get username and email from git config.
    Return empty if not configured or git is not found.
    """
    git = shutil.which("git")
    if not git:
        return "", ""
    try:
        username = subprocess.check_output([git, "config", "user.name"], text=True, encoding="utf-8").strip()
    except subprocess.CalledProcessError:
        username = ""
    try:
        email = subprocess.check_output([git, "config", "user.email"], text=True, encoding="utf-8").strip()
    except subprocess.CalledProcessError:
        email = ""
    return username, email


def add_ssh_scheme_to_git_uri(uri: str) -> str:
    """Cleans VCS uris from pip format"""
    # Add scheme for parsing purposes, this is also what pip does
    if "://" not in uri:
        uri = "ssh://" + uri
        parsed = parse.urlparse(uri)
        if ":" in parsed.netloc:
            netloc, _, path_start = parsed.netloc.rpartition(":")
            path = f"/{path_start}{parsed.path}"
            uri = parse.urlunparse(parsed._replace(netloc=netloc, path=path))
    return uri


@contextlib.contextmanager
def atomic_open_for_write(filename: str | Path, *, mode: str = "w", encoding: str = "utf-8") -> Iterator[IO]:
    dirname = os.path.dirname(filename)
    if not os.path.exists(dirname):
        os.makedirs(dirname)
    fd, name = tempfile.mkstemp(prefix="atomic-write-", dir=dirname)
    fp = open(fd, mode, encoding=encoding if "b" not in mode else None)
    try:
        yield fp
    except Exception:
        fp.close()
        raise
    else:
        fp.close()
        with contextlib.suppress(OSError):
            os.unlink(filename)
        # The tempfile is created with mode 600, we need to restore the default mode
        # with copyfile() instead of move().
        # See: https://github.com/pdm-project/pdm/issues/542
        shutil.copyfile(name, str(filename))
    finally:
        os.unlink(name)


@contextlib.contextmanager
def cd(path: str | Path) -> Iterator:
    _old_cwd = os.getcwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(_old_cwd)


def url_to_path(url: str) -> str:
    """
    Convert a file: URL to a path.
    """
    from urllib.request import url2pathname

    WINDOWS = sys.platform == "win32"

    if not url.startswith("file:"):
        raise ValueError(f"You can only turn file: urls into filenames (not {url!r})")

    _, netloc, path, _, _ = parse.urlsplit(url)

    if not netloc or netloc == "localhost":
        # According to RFC 8089, same as empty authority.
        netloc = ""
    elif WINDOWS:
        # If we have a UNC path, prepend UNC share notation.
        netloc = "\\\\" + netloc
    else:
        raise ValueError(f"non-local file URIs are not supported on this platform: {url!r}")

    path = url2pathname(netloc + path)

    # On Windows, urlsplit parses the path as something like "/C:/Users/foo".
    # This creates issues for path-related functions like io.open(), so we try
    # to detect and strip the leading slash.
    if (
        WINDOWS
        and not netloc  # Not UNC.
        and len(path) >= 3
        and path[0] == "/"  # Leading slash to strip.
        and path[1].isalpha()  # Drive letter.
        and path[2:4] in (":", ":/")  # Colon + end of string, or colon + absolute path.
    ):
        path = path[1:]

    return path


def split_path_fragments(path: Path) -> tuple[Path, str]:
    """Split a path into fragments"""
    left, sep, right = path.as_posix().partition("#egg=")
    return Path(left), sep + right


def expand_env_vars(credential: str, quote: bool = False, env: Mapping[str, str] | None = None) -> str:
    """A safe implementation of env var substitution.
    It only supports the following forms:

        ${ENV_VAR}

    Neither $ENV_VAR and %ENV_VAR is supported.
    """
    if env is None:
        env = os.environ

    def replace_func(match: Match) -> str:
        rv = env.get(match.group(1), "")
        return parse.quote(rv, "") if quote else rv

    return re.sub(r"\$\{(.+?)\}", replace_func, credential)


def expand_env_vars_in_auth(url: str) -> str:
    """In-place expand the auth in url"""
    scheme, netloc, path, params, query, fragment = parse.urlparse(url)
    if "@" in netloc:
        auth, rest = netloc.split("@", 1)
        auth = expand_env_vars(auth, True)
        netloc = "@".join([auth, rest])
    return parse.urlunparse((scheme, netloc, path, params, query, fragment))


@functools.lru_cache
def path_replace(pattern: str, replace_with: str, dest: str) -> str:
    """Safely replace the pattern in a path with given string.

    :param pattern: the pattern to match
    :param replace_with: the string to replace with
    :param dest: the path to replace
    :return the replaced path
    """
    sub_flags = re.IGNORECASE if os_name == "nt" else 0
    return re.sub(
        pattern.replace("\\", "/"),
        replace_with,
        dest.replace("\\", "/"),
        flags=sub_flags,
    )


def is_path_relative_to(path: str | Path, other: str | Path) -> bool:
    try:
        Path(path).relative_to(other)
    except ValueError:
        return False
    return True


def get_venv_like_prefix(interpreter: str | Path) -> tuple[Path | None, bool]:
    """Check if the given interpreter path is from a virtualenv,
    and return two values: the root path and whether it's a conda env.
    """
    interpreter = Path(interpreter)
    prefix = interpreter.parent
    if prefix.joinpath("conda-meta").exists():
        return prefix, True

    prefix = prefix.parent
    if prefix.joinpath("pyvenv.cfg").exists():
        return prefix, False
    if prefix.joinpath("conda-meta").exists():
        return prefix, True

    virtual_env = os.getenv("VIRTUAL_ENV")
    if virtual_env and is_path_relative_to(interpreter, virtual_env):
        return Path(virtual_env), False
    virtual_env = os.getenv("CONDA_PREFIX")
    if virtual_env and is_path_relative_to(interpreter, virtual_env):
        return Path(virtual_env), True
    return None, False


def find_python_in_path(path: str | Path) -> Path | None:
    """Find a python interpreter from the given path, the input argument could be:

    - A valid path to the interpreter
    - A Python root directory that contains the interpreter
    """
    pathlib_path = Path(path).absolute()
    if pathlib_path.is_file():
        return pathlib_path

    if os.name == "nt":
        for root_dir in (pathlib_path, pathlib_path / "Scripts"):
            if root_dir.joinpath("python.exe").exists():
                return root_dir.joinpath("python.exe")
    else:
        executable_pattern = re.compile(r"python(?:\d(?:\.\d+m?)?)?$")

        for python in pathlib_path.joinpath("bin").glob("python*"):
            if executable_pattern.match(python.name):
                return python

    return None


def get_rev_from_url(url: str) -> str:
    """Get the rev part from the VCS URL."""
    path = parse.urlparse(url).path
    if "@" in path:
        _, rev = path.rsplit("@", 1)
        return rev
    return ""


@functools.lru_cache
def normalize_name(name: str, lowercase: bool = True) -> str:
    name = re.sub(r"[^A-Za-z0-9]+", "-", name)
    return name.lower() if lowercase else name


def comparable_version(version: str) -> Version:
    """Normalize a version to make it valid in a specifier."""
    parsed = parse_version(version or "0.0.0")
    if parsed.local is not None:
        # strip the local part to make
        # comparable_version("1.2.3+local1") == Version("1.2.3")
        if hasattr(parsed, "__replace__"):  # packaging >= 26
            parsed = parsed.__replace__(local=None)
        else:
            # packaging < 26 does not have __replace__ method
            # In this version, we need to manually update _version and recompute _key
            # Note: In packaging >= 26, _key is a read-only property, but this else branch
            # only executes on packaging < 26 where _key is a regular attribute that can be
            # assigned. We use object.__setattr__() instead of direct assignment to satisfy
            # type checkers that analyze based on the current packaging version.
            from packaging.version import _cmpkey

            parsed._version = parsed._version._replace(local=None)

            object.__setattr__(
                parsed,
                "_key",
                _cmpkey(
                    parsed._version.epoch,
                    parsed._version.release,
                    parsed._version.pre,
                    parsed._version.post,
                    parsed._version.dev,
                    parsed._version.local,
                ),
            )

    return parsed


def is_egg_link(dist: Distribution) -> bool:
    """Check if the distribution is an egg-link install"""
    return getattr(dist, "link_file", None) is not None


def is_editable(dist: Distribution) -> bool:
    """Check if the distribution is installed in editable mode"""
    if is_egg_link(dist):
        return True
    direct_url = dist.read_text("direct_url.json")
    if not direct_url:
        return False
    direct_url_data = json.loads(direct_url)
    return direct_url_data.get("dir_info", {}).get("editable", False)


def pdm_scheme(base: str) -> dict[str, str]:
    """Return a PEP 582 style install scheme"""
    if "pep582" not in sysconfig.get_scheme_names():
        bin_prefix = "Scripts" if os.name == "nt" else "bin"
        sysconfig._INSTALL_SCHEMES["pep582"] = {  # type: ignore[attr-defined]
            "stdlib": "{pep582_base}/lib",
            "platstdlib": "{pep582_base}/lib",
            "purelib": "{pep582_base}/lib",
            "platlib": "{pep582_base}/lib",
            "include": "{pep582_base}/include",
            "scripts": f"{{pep582_base}}/{bin_prefix}",
            "data": "{pep582_base}",
            "prefix": "{pep582_base}",
            "headers": "{pep582_base}/include",
        }
    sysconfig.get_config_vars()["pep582_base"] = base
    return sysconfig.get_paths("pep582", vars={"pep582_base": base}, expand=True)


def is_url(url: str) -> bool:
    """Check if the given string is a URL"""
    return bool(parse.urlparse(url).scheme)


@functools.lru_cache
def fs_supports_link_method(method: str) -> bool:
    if not hasattr(os, method):
        return False
    if sys.platform == "win32":
        with tempfile.TemporaryDirectory(prefix="TmP") as temp_dir:
            with open(src := os.path.join(temp_dir, "a"), "w") as tmp_file:
                tmp_file.write("foo")
            dest = f"{src}-link"
            try:
                getattr(os, method)(src, dest)
                return True
            except (OSError, NotImplementedError):
                return False
    else:
        return True


def deprecation_warning(message: str, stacklevel: int = 1, raise_since: str | None = None) -> None:  # pragma: no cover
    """Show a deprecation warning with the given message and raise an error
    after a specified version.
    """
    from pdm.__version__ import __version__

    if raise_since is not None:
        if parse_version(__version__) >= parse_version(raise_since):
            raise PDMDeprecationWarning(message)
    warnings.warn(message, PDMDeprecationWarning, stacklevel=stacklevel + 1)


def is_pip_compatible_with_python(python_version: Version | str) -> bool:
    """Check the given python version is compatible with the pip installed"""
    from pdm.compat import importlib_metadata
    from pdm.models.specifiers import get_specifier

    pip = importlib_metadata.distribution("pip")
    requires_python = get_specifier(pip.metadata.get("Requires-Python"))
    return requires_python.contains(python_version, True)


def is_in_zipapp() -> bool:
    """Check if the current process is running in a zipapp"""
    return not os.path.exists(__file__)


@functools.lru_cache(None)
def package_installed(package_name: str) -> bool:
    try:
        importlib_metadata.distribution(package_name)
    except importlib_metadata.PackageNotFoundError:
        return False
    else:
        return True


def validate_project_name(name: str) -> bool:
    """Check if the project name is valid or not"""

    pattern = r"^([A-Z0-9]|[A-Z0-9][A-Z0-9._-]*[A-Z0-9])$"
    return re.fullmatch(pattern, name, flags=re.IGNORECASE) is not None


def sanitize_project_name(name: str) -> str:
    """Sanitize the project name and remove all illegal characters"""
    pattern = r"[^a-zA-Z0-9\-_\.]+"
    result = re.sub(pattern, "-", name)
    result = re.sub(r"^[\._-]|[\._-]$", "", result)
    if not result:
        raise PdmException(f"Invalid project name: {name}")
    return result


def is_conda_base() -> bool:
    return os.getenv("CONDA_DEFAULT_ENV", "") == "base"


def is_conda_base_python(python: Path) -> bool:
    if not is_conda_base():
        return False
    prefix = os.environ["CONDA_PREFIX"]
    try:
        python.relative_to(prefix)
    except ValueError:
        return False
    return True


def filtered_sources(sources: list[RepositoryConfig], package: str | None) -> list[RepositoryConfig]:
    """Get matching sources based on the index attribute."""
    source_preferences = [(s, _source_preference(package, s)) for s in sources]
    included_by = [s for s, p in source_preferences if p is True]
    if included_by:
        return included_by
    return [s for s, p in source_preferences if p is None]


def _source_preference(package: str | None, source: RepositoryConfig) -> bool | None:
    import fnmatch

    if package is None:
        return None
    key = normalize_name(package)
    if any(fnmatch.fnmatch(key, pat) for pat in source.include_packages):
        return True
    if any(fnmatch.fnmatch(key, pat) for pat in source.exclude_packages):
        return False
    return None


def get_file_hash(filename: str | Path, algorithm: str = "sha256") -> str:
    """Calculate the hash of a file with the given algorithm"""
    import hashlib

    h = hashlib.new(algorithm)
    with open(filename, "rb") as f:
        for chunk in iter(lambda: f.read(8192), b""):
            h.update(chunk)
    return h.hexdigest()


def convert_to_datetime(value: str) -> datetime:
    if match := re.fullmatch(r"(\d+)([dhw])", value):
        amount = int(match.group(1))
        unit = match.group(2)
        delta = {
            "d": timedelta(days=amount),
            "h": timedelta(hours=amount),
            "w": timedelta(weeks=amount),
        }[unit]
        return datetime.now(timezone.utc) - delta
    if "T" in value:
        return datetime.fromisoformat(value.replace("Z", "+00:00"))
    return datetime.strptime(value, "%Y-%m-%d").replace(tzinfo=timezone.utc)


def get_all_installable_python_versions(build_dir: bool = False) -> list[PythonVersion]:
    """Returns all installable standalone Python interpreter versions from @indygreg

    Installable means:
        Fitting current platform and arch

    Parameters:
        build_dir: Whether to include the `build/` directory from indygreg builds (aka 'Full Archive')
    """
    from pbs_installer._install import THIS_ARCH, THIS_PLATFORM
    from pbs_installer._versions import PYTHON_VERSIONS

    arch = "x86" if THIS_ARCH == "32" else THIS_ARCH
    matches = [v for v, u in PYTHON_VERSIONS.items() if any(k[:2] == (THIS_PLATFORM, arch) for k in u)]
    return matches


def get_class_init_params(klass: type) -> set[str]:
    arguments: set[str] = set()
    for cls in klass.__mro__:
        if "__init__" not in cls.__dict__:
            continue
        params = inspect.signature(cls).parameters
        arguments.update({k for k, v in params.items() if v.kind not in (v.VAR_POSITIONAL, v.VAR_KEYWORD)})
        if not any(p.kind in (p.VAR_POSITIONAL, p.VAR_KEYWORD) for p in params.values()):
            break
    return arguments


def get_requirement_from_override(name: str, value: str) -> str:
    if is_url(value):
        req = f"{name} @ {value}"
    else:
        try:
            SpecifierSet(value)
        except InvalidSpecifier:
            req = f"{name}=={value}"
        else:
            req = f"{name}{value}"
    return req


def hide_url(url: str) -> HiddenText:
    """Redact the URL for display purposes."""
    from pdm._types import HiddenText

    parsed = parse.urlsplit(url)
    if "@" not in parsed.netloc:
        return HiddenText(url, url)
    *_, netloc = parsed.netloc.rpartition("@")
    netloc = f"*****@{netloc}"
    redacted = parse.urlunsplit((parsed.scheme, netloc, parsed.path, parsed.query, parsed.fragment))
    return HiddenText(url, redacted)


def make_file_executable(path: str | Path) -> None:
    """Make the file at the provided path executable."""
    path_ = Path(path)
    path_.chmod(path_.stat().st_mode | 0o111)



from __future__ import annotations

import hashlib
import json
from typing import Any, cast

import tomlkit

from pdm import termui
from pdm.exceptions import ProjectError
from pdm.project.toml_file import TOMLFile
from pdm.utils import normalize_name


def _remove_empty_tables(doc: dict) -> None:
    for k, v in list(doc.items()):
        if isinstance(v, dict):
            _remove_empty_tables(v)
            if not v:
                del doc[k]


class PyProject(TOMLFile):
    """The data object representing th pyproject.toml file"""

    def _parse(self) -> dict[str, Any]:
        data = super()._parse()
        self._convert_pyproject(data)
        return data

    def open_for_write(self) -> tomlkit.TOMLDocument:
        if self._for_write:
            return cast(tomlkit.TOMLDocument, self._data)
        doc = super().open_for_write()
        self._convert_pyproject(doc)
        return doc

    def _convert_pyproject(self, data: dict[str, Any]) -> None:
        from pdm.formats import flit, poetry

        if "project" not in data and self._path.exists():
            # Try converting from flit and poetry
            for converter in (flit, poetry):
                if converter.check_fingerprint(None, self._path):
                    metadata, settings = converter.convert(None, self._path, None)
                    data["project"] = metadata
                    if settings:
                        data.setdefault("tool", {}).setdefault("pdm", {}).update(settings)
                    break

    def write(self, show_message: bool = True) -> None:
        """Write the TOMLDocument to the file."""
        _remove_empty_tables(self._data.get("project", {}))
        if "tool" in self._data:
            tool_table = cast(dict, self._data["tool"])
            _remove_empty_tables(tool_table.get("pdm", {}))
            if "pdm" in tool_table and not tool_table["pdm"]:
                del tool_table["pdm"]
            if not tool_table:
                del self._data["tool"]

        if "dependency-groups" in self._data and not self.dependency_groups:
            del self._data["dependency-groups"]
        super().write()
        if show_message:
            self.ui.echo("Changes are written to [success]pyproject.toml[/].", verbosity=termui.Verbosity.NORMAL)

    @property
    def is_valid(self) -> bool:
        return bool(self._data.get("project"))

    @property
    def metadata(self) -> dict[str, Any]:
        return self._data.setdefault("project", {})

    @property
    def dependency_groups(self) -> dict[str, Any]:
        return self._data.setdefault("dependency-groups", {})

    @property
    def dev_dependencies(self) -> dict[str, list[Any]]:
        groups: dict[str, list[Any]] = {}
        for group, deps in self._data.get("dependency-groups", {}).items():
            group = normalize_name(group)
            if group in groups:
                raise ProjectError(f"The group {group} is duplicated in dependency-groups")
            groups[group] = deps.unwrap() if hasattr(deps, "unwrap") else deps
        for group, deps in self.settings.get("dev-dependencies", {}).items():
            group = normalize_name(group)
            groups.setdefault(group, []).extend(deps.unwrap() if hasattr(deps, "unwrap") else deps)
        return groups

    @property
    def settings(self) -> dict[str, Any]:
        return self._data.setdefault("tool", {}).setdefault("pdm", {})

    @property
    def build_system(self) -> dict[str, Any]:
        return self._data.get("build-system", {})

    @property
    def resolution(self) -> dict[str, Any]:
        """A compatible getter method for the resolution overrides
        in the pyproject.toml file.
        """
        return self.settings.get("resolution", {})

    @property
    def allow_prereleases(self) -> bool | None:
        return self.resolution.get("allow-prereleases")

    def content_hash(self, algo: str = "sha256") -> str:
        """Generate a hash of the sensible content of the pyproject.toml file.
        When the hash changes, it means the project needs to be relocked.
        """
        dump_data = {
            "sources": self.settings.get("source", []),
            "dependencies": self.metadata.get("dependencies", []),
            "dev-dependencies": self.dev_dependencies,
            "optional-dependencies": self.metadata.get("optional-dependencies", {}),
            "requires-python": self.metadata.get("requires-python", ""),
            "resolution": self.resolution,
        }
        pyproject_content = json.dumps(dump_data, sort_keys=True)
        hasher = hashlib.new(algo)
        hasher.update(pyproject_content.encode("utf-8"))
        return hasher.hexdigest()

    @property
    def plugins(self) -> list[str]:
        return self.settings.get("plugins", [])



from __future__ import annotations

import os
import re
import shlex
from functools import cached_property
from pathlib import Path

from pdm.environments.base import BaseEnvironment
from pdm.utils import pdm_scheme


def _get_shebang_path(executable: str, is_launcher: bool) -> bytes:
    """Get the interpreter path in the shebang line

    The launcher can just use the command as-is.
    Otherwise if the path contains whitespace or is too long, both distlib
    and installer use a clever hack to make the shebang after ``/bin/sh``,
    where the interpreter path is quoted.
    """
    if is_launcher or (" " not in executable and (len(executable) + 3) <= 127):
        return executable.encode("utf-8")
    return shlex.quote(executable).encode("utf-8")


def _is_console_script(content: bytes) -> bool:
    import io
    import zipfile

    if os.name == "nt":  # Windows .exe should be a zip file.
        try:
            with zipfile.ZipFile(io.BytesIO(content)) as zf:
                return zf.namelist() == ["__main__.py"]
        except zipfile.BadZipFile:
            # Not a valid zip executable; fall back to text-based shebang detection below
            pass

    try:
        text = content.decode("utf-8")
        return text.startswith("#!")
    except UnicodeDecodeError:
        return False


def _replace_shebang(path: Path, new_executable: bytes) -> None:
    """Replace the python executable from the shebang line, which can be in two forms:

    1. #!python_executable
    2. #!/bin/sh
       '''exec' '/path to/python' "$0" "$@"
       ' '''
    """
    _complex_shebang_re = rb"^(#!/bin/sh\n'''exec' )('.+?')( \"\$0\")"
    _simple_shebang_re = rb"^(#!)(.+?)\s*(?=\n)"
    contents = path.read_bytes()

    if not _is_console_script(contents):
        return

    if os.name == "nt":
        new_content, count = re.subn(_simple_shebang_re, rb"\1" + new_executable, contents, count=1, flags=re.M)
        if count > 0:
            path.write_bytes(new_content)
        return

    new_content, count = re.subn(_complex_shebang_re, rb"\1" + new_executable + rb"\3", contents, count=1)
    if count > 0:
        path.write_bytes(new_content)
        return

    new_content, count = re.subn(_simple_shebang_re, rb"\1" + new_executable, contents, count=1)
    if count > 0:
        path.write_bytes(new_content)


class PythonLocalEnvironment(BaseEnvironment):
    """A project environment that installs packages into
    the local `__pypackages__` directory(PEP 582).
    """

    is_local = True

    @property
    def process_env(self) -> dict[str, str]:
        from pdm.cli.utils import get_pep582_path

        env = super().process_env
        pythonpath = os.getenv("PYTHONPATH", "").split(os.pathsep)
        pythonpath = [get_pep582_path(self.project)] + [p for p in pythonpath if "/pep582" not in p.replace("\\", "/")]
        env["PYTHONPATH"] = os.pathsep.join(pythonpath)
        env["PEP582_PACKAGES"] = str(self.packages_path)
        return env

    @cached_property
    def packages_path(self) -> Path:
        """The local packages path."""
        pypackages = self.project.root / "__pypackages__" / self.interpreter.identifier
        if not pypackages.exists() and "-32" in pypackages.name:
            compatible_packages = pypackages.with_name(pypackages.name[:-3])
            if compatible_packages.exists():
                pypackages = compatible_packages
        scripts = "Scripts" if os.name == "nt" else "bin"
        if not pypackages.parent.exists():
            pypackages.parent.mkdir(parents=True)
            pypackages.parent.joinpath(".gitignore").write_text("*\n!.gitignore\n")
        for subdir in [scripts, "include", "lib"]:
            pypackages.joinpath(subdir).mkdir(exist_ok=True, parents=True)
        return pypackages

    def get_paths(self, dist_name: str | None = None) -> dict[str, str]:
        scheme = pdm_scheme(self.packages_path.as_posix())
        scheme["headers"] = os.path.join(scheme["headers"], dist_name or "UNKNOWN")
        return scheme

    def update_shebangs(self, new_path: str) -> None:
        """Update the shebang lines"""
        scripts = self.get_paths()["scripts"]
        for child in Path(scripts).iterdir():
            if not child.is_file() or child.suffix not in (".exe", ".py", ""):
                continue
            is_launcher = child.suffix == ".exe"
            new_shebang = _get_shebang_path(new_path, is_launcher)
            _replace_shebang(child, new_shebang)
